#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from copy import deepcopy
from typing import Optional, Sequence, TypeVar, Union
import warnings

import casadi as ca
import numpy as np

from ..modules.base import Base, TimeSeries
from ..modules.controller.base import Controller
from ..util.util import is_list_like


Module = TypeVar('Module', bound=Base)
Control = TypeVar('Control', bound=Controller)
Numeric = Union[int, float]
NumArray = Union[Sequence[Numeric], np.ndarray]
NumNoneArray = Union[Sequence[Optional[Numeric]], np.ndarray]


_CHIRP_CHIRP = """
                                                  *#                                                    
                                             %##    %%%                                                 
                                          (%#       %%%%%#                                              
                                         %%     %%%       %%                                            
                                        %%    %%                                                        
                                       %#   %%                                                          
                                     %%%#**%(*%%%%                                                      
                                 %%%****//%#//////%%                                                    
                               %%****/////%/////////%%                                                  
                             %%***///////////////////%%                                                 
                           %%****/////////////////////%%                     %%(((%(                    
                          %%****////////((((((/////////%                 #%((###%%#%                    
                         %%****////////(  @@@@ (///////%              %%((####%% %##%*%%/**#%           
                         %****////////(( @  @@@(///////%           %%((#####%/   %%****//%%.            
                        /%***//////////(( @@@@////////%%         %#((#####%% %%/***///%%%/%             
                        ,%***////////////////////////%%%%%%,  %%((######%%%%****////%%  %/&             
                         %/***//////////////////////%###(((((%%(######%%%/***/////%%%   %%              
                          %%***///////%%/////////%%###((((((((%(%%%%%%****//////%%#%%  %*%              
                            %%****/%%///////%%%%####(((((((((%%(((%****///////%%#%#%%  %/%       /%%%%%%
                                    /#/,   %#####(((((((((((#%##%/***//////(%%((((((((%*/%((((#######%% 
                               /|           %####((((((((((%%##%****//////%%##########%*########%%%.    
                              / |            %#####(((((((%***%%**/////%%%############%/%#%%%%**%#      
   __________________________/  |             %%##%%###%%*****/%%**/%%%%%%%%%%%%%%%%%%*/%**//,*%#       
  /                             \            /%  %%%%%***,,////,,,,///****//*****/***%//%,,,/%%         
 |  chirp, chirp, chirp, ...     |          /%    %% .%%*///,,,,,,////,,,,//,,,,*//,,%//#,%%%           
  \_____________________________/          %%    /%     %%#,,,,,,////,,,,///,,,,///,,%///%              
                                         %%%     %%        %%%%(///,,,,,*///,,,,(#%%%%%//%              
                                       %%%%     %%               ###%%%%%%%%%%%###%%  %//%              
                                       #      %%%%                            %%%#####%//%              
                                            %%%%%                                     %///%             
                                            %%%                                       *%////////%%%  
"""


class DataSet:
    """"""
    def __init__(
            self,
            features: Sequence[str],
            labels: Sequence[str],
            add_time: bool = False,
            properties: Optional[dict[str, Union[Numeric, dict[str, Sequence[str]]]]] = None,
            plot_backend: Optional[str] = None
    ) -> None:
        """Constructor method"""
        self._raw_data = TimeSeries(backend=plot_backend, parent=self)
        self._train_data = TimeSeries(backend=plot_backend, parent=self)
        self._test_data = TimeSeries(backend=plot_backend, parent=self)

        n_features = len(features)
        n_labels = len(labels)

        vector = deepcopy(properties)
        if vector is None:
            vector = {}
        if add_time:
            if 'sampling_time' not in vector:
                vector['sampling_time'] = 1.
            if 'time' not in vector:
                vector['time'] = {}
        if 'x' not in vector:
            vector['x'] = {}
        if 'y' not in vector:
            vector['y'] = {}

        if add_time:
            vector['dt'] = vector.pop('sampling_time')
            vector['t'] = vector.pop('time')
            vec = vector['t']
            vec['values_or_names'] = ['t']
            if 'description' not in vec:
                vec['description'] = ['time...']
            if 'labels' not in vec:
                vec['labels'] = ['time']
            if 'units' not in vec:
                vec['units'] = ['h']
            vec['shape'] = (1, 0)
            vec['data_format'] = ca.DM

        vec = vector['x']
        vec['values_or_names'] = features
        if 'description' not in vec:
            vec['description'] = n_features * ['']
        if 'labels' not in vec:
            vec['labels'] = n_features * ['']
        if 'units' not in vec:
            vec['units'] = n_features * ['']
        vec['shape'] = (n_features, 0)
        vec['data_format'] = ca.DM

        vec = vector['y']
        vec['values_or_names'] = labels
        if 'description' not in vec:
            vec['description'] = n_labels * ['']
        if 'labels' not in vec:
            vec['labels'] = n_labels * ['']
        if 'units' not in vec:
            vec['units'] = n_labels * ['']
        vec['shape'] = (n_labels, 0)
        vec['data_format'] = ca.DM

        names = vector.keys()

        self._raw_data.setup(*names, **vector)
        self._train_data.setup(*names, **vector)
        self._test_data.setup(*names, **vector)
        self._train_index = []
        self._test_index = []
        self._x_noise_added = False
        self._y_noise_added = False

    def __len__(self) -> int:
        """Length method"""
        return self._raw_data.n_samples

    def _reduce_data(
            self,
            method: str,
            distance_threshold: Optional[Numeric] = None,
            downsample_factor: Optional[int] = None
    ) -> (int, NumArray):
        """

        :param method:
        :param distance_threshold:
        :param downsample_factor:
        :return:
        """
        if self._raw_data.is_empty():
            raise ValueError("No raw data available")

        data_removal = method.lower().replace(' ', '_')
        if data_removal == 'euclidean_distance' and distance_threshold is None:
            warnings.warn("No distance threshold supplied for data selection using Euclidean distance. "
                          "Applying default value of 0.5.")
            distance_threshold = .5
        if data_removal == 'downsample' and downsample_factor is None:
            warnings.warn("No downsample factor supplied for data selection using downsampling. "
                          "Applying default value of 10.")
            downsample_factor = 10

        inputs = self._raw_data.get_by_id('x').full()
        index = np.array([])

        k = 0
        dim = inputs.shape[1]
        n_inputs = inputs.shape[1]
        while True:
            if data_removal == 'euclidean_distance':
                euclidean_distance = np.zeros(n_inputs)
                euclidean_distance[:k + 1] = distance_threshold * np.ones(k + 1)

                distance = inputs[:, None, k] - inputs[:, k + 1:]
                k += 1
                if distance.size == 0:
                    break

                euclidean_distance[k:] = np.linalg.norm(distance, axis=0)

                index_keep = euclidean_distance >= distance_threshold
                if index.size == 0:
                    index = np.flatnonzero(index_keep)
                else:
                    index = index[index_keep]
            elif data_removal == 'downsample':
                index = np.arange(0, inputs.shape[1], downsample_factor)
                break
            else:
                raise NotImplementedError(f"Data selection method '{method}' not implemented or recognized")

            inputs = inputs[:, index_keep]

            n_inputs = inputs.shape[1]
            if n_inputs <= k:
                break

        return dim, index

    def _plot_selected_data(self, label: str, index: NumArray, *args, **kwargs):
        """

        :param label:
        :param args:
        :param kwargs:
        :return:
        """
        x_data = self._raw_data.to_dict(*[arg[0] for arg in args], subplots=True, suffix=label, index=index)
        x_data = [value for value in x_data.values()]
        y_data = self._raw_data.to_dict(*[arg[1] for arg in args], subplots=True, suffix=label, index=index)
        y_data_keys = y_data.keys()
        y_data = [value for value in y_data.values()]
        for key, label in enumerate(y_data_keys):
            x_data[key]['label'] = label
            y_data[key]['label'] = label
            y_data[key]['kind'] = 'scatter'

        plot_kwargs = kwargs.copy()
        plot_kwargs['marker'] = kwargs.get('marker', 'o')
        plot_kwargs['marker_size'] = kwargs.get('marker_size')
        if plot_kwargs['marker_size'] is None:
            if self._raw_data.plot_backend == 'bokeh':
                plot_kwargs['marker_size'] = 5
            elif self._raw_data.plot_backend == 'matplotlib':
                plot_kwargs['marker_size'] = 20

        return self.plot_raw_data(*args, x_data=x_data, y_data=y_data, **plot_kwargs)

    @property
    def features(self) -> list[str]:
        """

        :return:
        """
        return self._raw_data.get_names('x')

    @property
    def labels(self) -> list[str]:
        """

        :return:
        """
        return self._raw_data.get_names('y')

    @property
    def raw_data(self) -> (NumArray, NumArray):
        """

        :return:
        """
        if self._x_noise_added:
            feature_key = 'x_noisy'
        else:
            feature_key = 'x'
        if self._y_noise_added:
            label_key = 'y_noisy'
        else:
            label_key = 'y'
        features = self._raw_data.get_by_id(feature_key).full()
        labels = self._raw_data.get_by_id(label_key).full()

        return features, labels

    @property
    def train_data(self) -> (NumArray, NumArray):
        """

        :return:
        """
        if self._x_noise_added:
            feature_key = 'x_noisy'
        else:
            feature_key = 'x'
        if self._y_noise_added:
            label_key = 'y_noisy'
        else:
            label_key = 'y'
        features = self._raw_data.get_by_id(feature_key).full()
        labels = self._raw_data.get_by_id(label_key).full()

        return features[:, self._train_index], labels[:, self._train_index]

    @property
    def test_data(self) -> (NumArray, NumArray):
        """

        :return:
        """
        if self._x_noise_added:
            feature_key = 'x_noisy'
        else:
            feature_key = 'x'
        if self._y_noise_added:
            label_key = 'y_noisy'
        else:
            label_key = 'y'
        features = self._raw_data.get_by_id(feature_key).full()
        labels = self._raw_data.get_by_id(label_key).full()

        return features[:, self._test_index], labels[:, self._test_index]

    @property
    def sampling_time(self) -> Optional[float]:
        """

        :return:
        """
        return self._raw_data.dt

    dt = sampling_time

    @property
    def time_unit(self) -> Optional[str]:
        """

        :return:
        """
        if 't' in self._raw_data:
            return self._raw_data.get_units('t')
        return None

    def add_data(
            self,
            features: NumArray,
            labels: NumArray,
            time: Optional[NumArray] = None,
            feature_noise: Optional[NumArray] = None,
            label_noise: Optional[NumArray] = None
    ) -> None:
        """

        :param features:
        :param labels:
        :param time:
        :param feature_noise:
        :param label_noise:
        :return:
        """
        self._raw_data.add('x', features)
        self._raw_data.add('y', labels)
        if time is not None:
            if 't' in self._raw_data:
                self._raw_data.add('t', time)
            else:
                warnings.warn("No data array was set up for the time... No changes applied with respect to the time "
                              "vector")
        if feature_noise is not None:
            if not self._x_noise_added:
                self._x_noise_added = True
            self._raw_data.add('x_noise', feature_noise)
        if label_noise is not None:
            if not self._y_noise_added:
                self._y_noise_added = True
            self._raw_data.add('y_noise', label_noise)

    def set_data(
            self,
            features: NumArray,
            labels: NumArray,
            time: Optional[NumArray] = None,
            feature_noise: Optional[NumArray] = None,
            label_noise: Optional[NumArray] = None
    ) -> None:
        """

        :param features:
        :param labels:
        :param time:
        :param feature_noise:
        :param label_noise:
        :return:
        """
        self._raw_data.set('x', features)
        self._raw_data.set('y', labels)
        if time is not None:
            if 't' in self._raw_data:
                self._raw_data.set('t', time)
            else:
                warnings.warn("No data array was set up for the time... No changes applied with respect to the time "
                              "vector")
        if feature_noise is not None:
            if not self._x_noise_added:
                self._x_noise_added = True
            self._raw_data.set('x_noise', feature_noise)
        if label_noise is not None:
            if not self._y_noise_added:
                self._y_noise_added = True
            self._raw_data.set('y_noise', label_noise)

    def add_noise(
            self,
            *args,
            distribution: Union[str, Sequence[str]] = 'normal',
            seed: Optional[int] = None,
            **kwargs
    ) -> None:
        """

        :param args:
        :param distribution:
        :param seed:
        :param kwargs:
        :return:
        """
        if not self._x_noise_added:
            self._x_noise_added = True
        if not self._y_noise_added:
            self._y_noise_added = True
        self._raw_data.make_some_noise(*args, distribution=distribution, seed=seed, **kwargs)

    def select_train_data(
            self,
            method: str,
            distance_threshold: Optional[Numeric] = None,
            downsample_factor: Optional[int] = None
    ) -> None:
        """

        :param method:
        :param distance_threshold:
        :param downsample_factor:
        :return:
        """
        dim, index = self._reduce_data(method, distance_threshold=distance_threshold,
                                       downsample_factor=downsample_factor)
        self._train_index = index

        print(f"{len(index)}/{dim} data points selected for training")

    def select_test_data(
            self,
            method: str,
            distance_threshold: Optional[Numeric] = None,
            downsample_factor: Optional[int] = None
    ) -> None:
        """

        :param method:
        :param distance_threshold:
        :param downsample_factor:
        :return:
        """
        dim, index = self._reduce_data(method, distance_threshold=distance_threshold,
                                       downsample_factor=downsample_factor)
        self._test_index = index

        print(f"{len(index)}/{dim} data points selected for testing")

    def plot_raw_data(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        plot_kwargs = kwargs.copy()
        if self._raw_data.plot_backend == 'bokeh':
            plot_kwargs["line_width"] = kwargs.get("line_width", 2)
        elif self._raw_data.plot_backend == 'matplotlib':
            plot_kwargs["line_width"] = kwargs.get("line_width", 1)
        # NOTE: The following 2 lines will be ignored for backend 'matplotlib'
        plot_kwargs["major_label_text_font_size"] = kwargs.get("major_label_text_font_size", "12pt")
        plot_kwargs["axis_label_text_font_size"] = kwargs.get("axis_label_text_font_size", "12pt")
        return self._raw_data.plot(*args, **plot_kwargs)

    def plot_train_data(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        self._plot_selected_data('train', self._train_index, *args, **kwargs)

    def plot_test_data(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        self._plot_selected_data('test', self._test_index, *args, **kwargs)

    def copy(self, ignore_time: bool = False) -> 'DataSet':
        """

        :param ignore_time:
        :return:
        """
        add_time = 't' in self._raw_data and not self._raw_data['t'].is_empty() and not ignore_time
        if add_time:
            properties = {
                'sampling_time': self._raw_data.dt,
                'time': {
                    'description': [self._raw_data.get_description('t')],
                    'labels': [self._raw_data.get_labels('t')],
                    'units': [self._raw_data.get_units('t')]
                }
            }
        else:
            properties = {}
        properties['x'] = {
            'description': self._raw_data.get_description('x'),
            'labels': self._raw_data.get_labels('x'),
            'units': self._raw_data.get_units('x')
        }
        properties['y'] = {
            'description': self._raw_data.get_description('y'),
            'labels': self._raw_data.get_labels('y'),
            'units': self._raw_data.get_units('y')
        }
        data_set = DataSet(self._raw_data.get_names('x'), self._raw_data.get_names('y'), add_time=add_time,
                           properties=properties, plot_backend=self._raw_data.plot_backend)

        features = self._raw_data.get_by_id('x')
        labels = self._raw_data.get_by_id('y')
        kwargs = {}
        if add_time:
            kwargs['time'] = self._raw_data.get_by_id('t')
        feature_noise = self._raw_data.get_by_id('x_noise')
        if not feature_noise.is_empty():
            kwargs['feature_noise'] = feature_noise
        label_noise = self._raw_data.get_by_id('y_noise')
        if not label_noise.is_empty():
            kwargs['label_noise'] = label_noise

        data_set.set_data(features, labels, **kwargs)

        return data_set

    def sort(self, arg: str, order: str = 'descending') -> None:
        """

        :param arg:
        :param order:
        :return:
        """
        data = self._raw_data.get_by_name(arg)
        if data is not None and not data.is_empty():
            idx = np.argsort(data, axis=None)
            if order == 'descending':
                idx = idx[::-1]
            elif order != 'ascending':
                raise ValueError(f"Keyword argument order='{order}' not recognized")

            for arg in self._raw_data:
                data = self._raw_data.get_by_id(arg)
                noise = self._raw_data.get_by_id(arg + '_noise')
                self._raw_data.set(arg, data[:, idx])
                if not noise.is_empty():
                    self._raw_data.set(arg + '_noise', noise)

    def append(self, other: list['DataSet'], ignore_index: bool = False, sort: bool = True) -> 'DataSet':
        """

        :param other:
        :param ignore_index:
        :param sort:
        :return:
        """
        new_data_set = self.copy(ignore_time=ignore_index)

        # TODO: Add support for pandas objects
        if not ignore_index:
            dt = new_data_set._raw_data.dt
            time_unit = new_data_set._raw_data.get_units('t')
        else:
            dt = None
            time_unit = None
        features = new_data_set._raw_data.get_names('x')
        labels = new_data_set._raw_data.get_names('y')
        for data_set in other:
            if not ignore_index:
                other_dt = data_set.dt
                other_time_unit = data_set.time_unit
                if dt != other_dt or time_unit != other_time_unit:
                    warnings.warn(f"Different sampling times for supplied data sets. The data set to be appended has a "
                                  f"sampling time of dt='{other_dt} {other_time_unit}', but the data set to be extended"
                                  f" has a sampling time of dt='{dt} {time_unit}'. If time information is not required "
                                  f"in your case, ignore this message or set the flag ignore_index to True to prevent "
                                  f"the message from being shown in future.")
                if (other_dt is None and dt is not None) or (other_time_unit is None and time_unit is not None):
                    warnings.warn('An ambiguous data set with respect to time was supplied')
            if features != data_set.features:
                # TODO: Sort features of other data set, if just the order is different
                raise ValueError(f"Mismatch in the features. Got {data_set.features}, expected {features}.")
            if labels != data_set.labels:
                # TODO: Sort labels of other data set, if just the order is different
                raise ValueError(f"Mismatch in the labels. Got {data_set.labels}, expected {labels}.")

            # TODO: What to do here, when training data selection was executed? Do we just ignore it?
            other_features = data_set._raw_data.get_by_id('x')
            other_labels = data_set._raw_data.get_by_id('y')
            other_kwargs = {}
            if not ignore_index:
                other_t = data_set._raw_data.get_by_id('t')
                if not other_t.is_empty():
                    other_kwargs['time'] = other_t
            feature_noise = new_data_set._raw_data.get_by_id('x_noise')
            other_feature_noise = data_set._raw_data.get_by_id('x_noise')
            if not feature_noise.is_empty() and not other_feature_noise.is_empty():
                other_kwargs['feature_noise'] = other_feature_noise
            label_noise = new_data_set._raw_data.get_by_id('y_noise')
            other_label_noise = data_set._raw_data.get_by_id('y_noise')
            if not label_noise.is_empty() and not other_label_noise.is_empty():
                other_kwargs['label_noise'] = other_label_noise

            # NOTE: We ignore description, labels and units of the other data sets for now, since they should be the
            #  same ideally.
            new_data_set.add_data(other_features, other_labels, **other_kwargs)

        if sort:
            new_data_set.sort('t', order='ascending')

        return new_data_set


class DataGenerator:
    """"""
    def __init__(
            self,
            module: Module,
            x0: Optional[Union[Numeric, NumArray]] = None,
            z0: Optional[Union[Numeric, NumArray]] = None,
            p0: Optional[Union[Numeric, NumArray]] = None,
            use_input_as_label: bool = False
    ) -> None:
        """Constructor method"""
        self._module = module.copy(setup=True)

        if x0 is None:
            x0 = module.x0
            if x0 is None:
                raise RuntimeError("No initial dynamical states found. Please supply initial dynamical states or set "
                                   "initial conditions of the model before generating data!")

        if z0 is None:
            z0 = module.z0
            if module.n_z > 0 and z0 is None:
                raise RuntimeError("No initial algebraic states found. Please supply initial algebraic states or set "
                                   "initial conditions of the model before generating data!")

        if p0 is None:
            if 'p' not in module.solution or module.solution.get_by_id('p').is_empty():
                p0 = None
            else:
                p0 = module.solution.get_by_id('p:0')
            if module.n_p > 0 and p0 is None:
                raise RuntimeError("No parameter values found. Please supply parameter values or set values for the "
                                   "parameters of the model before generating data!")

        self._module.set_initial_conditions(x0, z0=z0)
        if p0 is not None:
            self._module.set_initial_parameter_values(p0)

        self._n_inputs = self._module.n_u
        self._samples = None
        self._control_loop = None
        self._use_input_as_label = use_input_as_label
        self._data_set = None

    @staticmethod
    def _linear_chirp(
            t: NumArray,
            chirp_rate: Numeric,
            initial_phase: Optional[Numeric] = None,
            initial_frequency: Optional[Numeric] = None
    ) -> np.ndarray:
        """

        :param t:
        :param chirp_rate:
        :param initial_phase:
        :param initial_frequency:
        :return:
        """
        if initial_phase is None:
            initial_phase = np.pi / 2.
        if initial_frequency is None:
            initial_frequency = .0001

        return np.sin(initial_phase + 2. * np.pi * (chirp_rate / 2. * t + initial_frequency) * t)

    @staticmethod
    def _exponential_chirp(
            t: NumArray,
            chirp_rate: Numeric,
            initial_phase: Optional[Numeric] = None,
            initial_frequency: Optional[Numeric] = None
    ) -> np.ndarray:
        """

        :param t:
        :param chirp_rate:
        :param initial_phase:
        :param initial_frequency:
        :return:
        """
        # TODO: Should we use other values here?
        if initial_phase is None:
            initial_phase = np.pi / 2.
        if initial_frequency is None:
            initial_frequency = .0001

        return np.sin(initial_phase + 2. * np.pi * initial_frequency * (chirp_rate ** t - 1) / np.log(chirp_rate))

    @staticmethod
    def _hyperbolic_chirp(
            t: NumArray,
            dt: Numeric,
            initial_frequency_ratio: Numeric,
            initial_phase: Optional[Numeric] = None,
            initial_frequency: Optional[Numeric] = None
    ) -> np.ndarray:
        """

        :param t:
        :param dt:
        :param initial_frequency_ratio:
        :param initial_phase:
        :param initial_frequency:
        :return:
        """
        # TODO: Should we use other values here?
        if initial_phase is None:
            initial_phase = np.pi / 2.
        if initial_frequency is None:
            initial_frequency = .0001

        fraction = 1. - 1. / initial_frequency_ratio
        return np.sin(initial_phase - 2. * np.pi * dt * initial_frequency / fraction * np.log(1. - fraction / dt * t))

    def _add_noise(self, index: NumArray, shape: tuple[int, int], **kwargs) -> None:
        """

        :param index:
        :param shape:
        :param kwargs:
        :return:
        """
        noise = np.zeros(shape)
        noise_added = False
        for k, x_name in enumerate(self._module.dynamical_state_names):
            if k not in index:
                continue
            x_noise = kwargs.get(x_name)
            if x_noise is not None:
                distribution, seed, info = _get_distribution_information(**x_noise)
                if seed is not None:
                    np.random.seed(seed)
                if distribution == 'random_uniform':
                    noise[index[k], :] = np.random.uniform(low=info[0], high=info[1], size=(1, shape[1]))
                elif distribution == 'random_normal':
                    noise[index[k], :] = np.random.normal(loc=0., scale=info, size=(1, shape[1]))
                else:
                    raise ValueError(f"Distribution '{distribution}' not available/recognized for adding noise to "
                                     f"generated data")
                if not noise_added:
                    noise_added = True
        if not noise_added:
            distribution, seed, info = _get_distribution_information(**kwargs)
            if seed is not None:
                np.random.seed(seed)
            if distribution == 'random_uniform':
                noise[index, :] = np.random.uniform(low=info[0], high=info[1], size=(len(index), shape[1]))
            elif distribution == 'random_normal':
                noise[index, :] = np.random.normal(loc=0., scale=info, size=(len(index), shape[1]))
            else:
                raise ValueError(f"Distribution '{distribution}' not available/recognized for adding noise to "
                                 f"generated data")

        self._samples = noise

    @property
    def data(self) -> Optional[DataSet]:
        """

        :return:
        """
        return self._data_set

    def random_uniform(
            self,
            n_samples: int,
            steps: int,
            lower_bound: Union[Numeric, NumArray],
            upper_bound: Union[Numeric, NumArray],
            seed: Optional[int] = None
    ) -> None:
        """

        :param n_samples:
        :param steps:
        :param lower_bound:
        :param upper_bound:
        :param seed:
        :return:
        """
        if seed is not None:
            np.random.seed(seed)

        samples = np.random.uniform(low=lower_bound, high=upper_bound, size=(n_samples, self._n_inputs))
        samples = np.repeat(samples, steps, axis=0)

        self._samples = samples.T

    def random_normal(
            self,
            n_samples: int,
            steps: int,
            mean: Union[Numeric, NumArray],
            variance: Union[Numeric, NumArray],
            seed: Optional[int] = None
    ) -> None:
        """

        :param n_samples:
        :param steps:
        :param mean:
        :param variance:
        :param seed:
        :return:
        """
        if seed is not None:
            np.random.seed(seed)

        samples = np.random.normal(loc=mean, scale=np.sqrt(variance), size=(n_samples, self._n_inputs))
        samples = np.repeat(samples, steps, axis=0)

        self._samples = samples.T

    def chirp(
            self,
            type_: Union[str, Sequence[str]],
            amplitude: Union[Numeric, NumArray],
            length: Union[Numeric, NumArray],
            mean: Union[Numeric, NumArray],
            chirp_rate: Union[Numeric, NumArray],
            initial_phase: Optional[Union[Numeric, NumNoneArray]] = None,
            initial_frequency: Optional[Union[Numeric, NumNoneArray]] = None,
            initial_frequency_ratio: Optional[Union[Numeric, NumArray]] = None
    ) -> None:
        """

        :param type_:
        :param amplitude:
        :param length:
        :param mean:
        :param chirp_rate:
        :param initial_phase:
        :param initial_frequency:
        :param initial_frequency_ratio:
        :return:
        """
        print(_CHIRP_CHIRP)

        if not is_list_like(type_):
            type_ = [[type_] for _ in range(self._n_inputs)]
        if not is_list_like(amplitude):
            amplitude = [[amplitude] for _ in range(self._n_inputs)]
        if not is_list_like(length):
            length = [[length] for _ in range(self._n_inputs)]
        if not is_list_like(mean):
            mean = [[mean] for _ in range(self._n_inputs)]
        if not is_list_like(chirp_rate):
            chirp_rate = [[chirp_rate] for _ in range(self._n_inputs)]
        if not is_list_like(initial_phase):
            initial_phase = [[initial_phase] for _ in range(self._n_inputs)]
        if not is_list_like(initial_frequency):
            initial_frequency = [[initial_frequency] for _ in range(self._n_inputs)]
        if not is_list_like(initial_frequency_ratio):
            initial_frequency_ratio = [[initial_frequency_ratio] for _ in range(self._n_inputs)]

        dt = self._module.solution.dt
        samples = []
        for i in range(self._n_inputs):
            ti = type_[i]
            ai = amplitude[i]
            li = length[i]
            mi = mean[i]
            ci = chirp_rate[i]
            phi = initial_phase[i]
            fri = initial_frequency[i]
            rati = initial_frequency_ratio[i]

            if not is_list_like(ti):
                ti = [ti]
            if not is_list_like(ai):
                ai = [ai]
            if not is_list_like(li):
                li = [li]
            if not is_list_like(mi):
                mi = [mi]
            if not is_list_like(ci):
                ci = [ci]
            if not is_list_like(phi):
                phi = [phi]
            if not is_list_like(fri):
                fri = [fri]
            if not is_list_like(rati):
                rati = [rati]

            lens = [len(ti), len(ai), len(li), len(mi), len(ci), len(phi), len(fri), len(rati)]
            n_chirps = max(lens)
            if any(k != n_chirps and k != 1 for k in lens):
                mismatch = []
                if lens[0] != 1:
                    mismatch.append(f'types ({lens[0]})')
                if lens[1] != 1:
                    mismatch.append(f'amplitudes ({lens[1]})')
                if lens[2] != 1:
                    mismatch.append(f'lengths ({lens[2]})')
                if lens[3] != 1:
                    mismatch.append(f'means ({lens[3]})')
                if lens[4] != 1:
                    mismatch.append(f'chirp rates ({lens[4]})')
                if lens[5] != 1:
                    mismatch.append(f'initial phases ({lens[5]})')
                if lens[6] != 1:
                    mismatch.append(f'initial frequencies ({lens[6]})')
                if lens[7] != 1:
                    mismatch.append(f'initial frequency ratios ({lens[7]})')
                raise ValueError(f"Dimension mismatch between {', '.join(mismatch[:-1])} and {mismatch[-1]}")

            if lens[0] != n_chirps:
                ti *= n_chirps
            if lens[1] != n_chirps:
                ai *= n_chirps
            if lens[2] != n_chirps:
                li *= n_chirps
            if lens[3] != n_chirps:
                mi *= n_chirps
            if lens[4] != n_chirps:
                ci *= n_chirps
            if lens[5] != n_chirps:
                phi *= n_chirps
            if lens[6] != n_chirps:
                fri *= n_chirps
            if lens[7] != n_chirps:
                rati *= n_chirps

            ui = []
            for j in range(n_chirps):
                t = np.arange(0., li[j], dt)

                tij = ti[j].lower()
                if tij == 'linear':
                    chirp = self._linear_chirp(t, ci[j], initial_phase=phi[j], initial_frequency=fri[j])
                elif tij == 'exponential':
                    # TODO: Test this
                    chirp = self._exponential_chirp(t, ci[j], initial_phase=phi[j], initial_frequency=fri[j])
                elif tij == 'hyperbolic':
                    # TODO: Test this
                    ratij = rati[j]
                    if ratij is None:
                        raise ValueError("Initial frequency ratio for hyperbolic chirp signal was not supplied")
                    chirp = self._hyperbolic_chirp(t, dt, ratij, initial_phase=phi[j], initial_frequency=fri[j])
                else:
                    raise ValueError(f"Type '{ti[j]}' not recognized for chirp signal")

                signal = mi[j] + ai[j] * chirp
                ui.append(signal)

            ui = np.concatenate(ui)
            samples.append(ui)

        samples = np.concatenate([samples], axis=1)

        self._samples = samples

    def closed_loop(self, controller: Control, steps: int) -> None:
        """

        :param controller:
        :param steps:
        :return:
        """
        controller_is_mpc = controller.type == 'NMPC'

        def run() -> None:
            """

            :return:
            """
            self._module.reset_solution()
            solution = self._module.solution
            x0 = solution.get_by_id('x:0')
            for _ in range(steps):
                if controller_is_mpc:
                    state_names = controller._model_orig.dynamical_state_names
                    ind_states = [self._module.dynamical_state_names.index(name) for name in state_names]

                    u = controller.optimize(x0[ind_states])
                self._module.simulate(u=u)
                x0 = solution.get_by_id('x:f')

        self._samples = steps
        self._control_loop = run

    def run(
            self,
            output: str,
            skip: Optional[Sequence[int]] = None,
            shift: int = 0.,
            add_noise: Optional[Union[dict[str, Numeric], dict[str, dict[str, Numeric]]]] = None
    ) -> None:
        """

        :param output:
        :param skip:
        :param shift:
        :param add_noise:
        :return:
        """
        # TODO: Support for algebraic states and parameters
        if self._control_loop is not None:
            self._control_loop()
            n_data_points = self._samples
        else:
            self._module.reset_solution()
            n_data_points = self._samples.shape[1]
            self._module.simulate(u=self._samples, steps=n_data_points)

        if skip is None:
            skip = []
        keep = [k for k in range(self._module.n_x) if k not in skip]

        t = self._module.solution.get_by_id('t')
        x = self._module.solution.get_by_id('x')
        u = self._module.solution.get_by_id('u')

        time = t[shift + 1:].full()
        if not self._use_input_as_label:
            if output == 'difference':
                outputs = np.diff(x, axis=1)[keep, shift:]
                features = [name for index, name in enumerate(self._module.dynamical_state_names) if
                            index in keep] + self._module.input_names
                labels = ['delta_' + name for index, name in enumerate(self._module.dynamical_state_names) if
                          index in keep]

                label_description = ['difference of ' + text for index, text in
                                     enumerate(self._module.dynamical_state_description) if index in keep]
            else:  # output == 'absolute'
                outputs = x[keep, shift + 1:].full()
                features = [name + '_k' for index, name in enumerate(self._module.dynamical_state_names) if
                            index in keep] + self._module.input_names
                labels = [name + '_k+1' for index, name in enumerate(self._module.dynamical_state_names) if
                          index in keep]

                label_description = [text for index, text in enumerate(self._module.dynamical_state_description) if
                                     index in keep]
            inputs = np.concatenate([x[keep, k:-(shift + 1 - k)] for k in range(shift + 1)] + [u[:, shift:]], axis=0)

            feature_description = [text for index, text in enumerate(self._module.dynamical_state_description) if
                                   index in keep] + self._module.input_description
            feature_labels = [text for index, text in enumerate(self._module.dynamical_state_labels) if
                              index in keep] + self._module.input_labels
            feature_units = [text for index, text in enumerate(self._module.dynamical_state_units) if
                             index in keep] + self._module.input_units

            label_labels = [text for index, text in enumerate(self._module.dynamical_state_labels) if index in keep]
            label_units = [text for index, text in enumerate(self._module.dynamical_state_units) if index in keep]
        else:
            if output == 'difference':
                warnings.warn("The behavior of choosing the difference of the inputs as labels has not been tested yet."
                              " So strange things can happen. Should you have an example where the difference of the "
                              "inputs as labels is required, we would appreciate it if you sent us that example so we "
                              "can refine this case.")
                outputs = np.diff(u, axis=1)[:, shift:]
                labels = ['delta_' + name for name in self._module.input_names]

                label_description = ['difference of ' + text for text in self._module.input_description]
            else:  # output == 'absolute'
                outputs = u[:, shift:].full()
                labels = self._module.input_names

                label_description = self._module.input_description
            inputs = np.concatenate([x[keep, k:-(shift + 1 - k)] for k in range(shift + 1)], axis=0)
            features = [name for index, name in enumerate(self._module.dynamical_state_names) if index in keep]

            feature_description = [text for index, text in enumerate(self._module.dynamical_state_description) if
                                   index in keep]
            feature_labels = [text for index, text in enumerate(self._module.dynamical_state_labels) if index in keep]
            feature_units = [text for index, text in enumerate(self._module.dynamical_state_units) if index in keep]

            label_labels = self._module.input_labels
            label_units = self._module.input_units

        if shift > 0:
            if output == 'difference':
                features = [name + '_k' if name in self._module.dynamical_state_names else name for name in features]

                feature_description = [
                    text + ' at time point k' if text and text in self._module.dynamical_state_description else text for
                    text in feature_description]

            for k in range(shift):
                features = [name + f'_k-{k + 1}' for index, name in enumerate(self._module.dynamical_state_names) if
                            index in keep] + features

                feature_description = [text + f' at time point k-{k + 1}' if text else text for index, text in
                                       enumerate(self._module.dynamical_state_description) if
                                       index in keep] + feature_description
                feature_labels = [text for index, text in enumerate(self._module.dynamical_state_labels) if
                                  index in keep] + feature_labels
                feature_units = [text for index, text in enumerate(self._module.dynamical_state_units) if
                                 index in keep] + feature_units

        if add_noise is not None:
            self._add_noise(keep, x.shape, **add_noise)
            noise = self._samples[keep, :]
            if output == 'difference':
                output_noise = np.diff(noise, axis=1)[:, shift:]
            else:  # output == 'absolute'
                output_noise = noise[:, shift + 1:]
            input_noise = np.concatenate([noise[:, k:-(shift + 1 - k)] for k in range(shift + 1)] + [
                np.zeros((self._module.n_u, n_data_points - shift))], axis=0)
        else:
            output_noise = None
            input_noise = None

        properties = {
            'sampling_time': self._module.solution.dt,
            'time': {
                'description': ['time...'],
                'labels': ['time'],
                'units': [self._module.time_unit]
            },
            'x': {
                'description': feature_description,
                'labels': feature_labels,
                'units': feature_units
            },
            'y': {
                'description': label_description,
                'labels': label_labels,
                'units': label_units
            }
        }

        self._data_set = DataSet(features, labels, add_time=True, properties=properties,
                                 plot_backend=self._module.solution.plot_backend)
        self._data_set.set_data(inputs, outputs, time=time, feature_noise=input_noise, label_noise=output_noise)


def _get_distribution_information(**kwargs) -> (str, Optional[int], Optional[Union[Numeric, NumArray]]):
    """

    :param kwargs:
    :return:
    """
    distribution = kwargs.get('distribution')
    if distribution is None:
        distribution = 'random_normal'

    seed = kwargs.get('seed')
    info = None

    if distribution == 'random_uniform':
        lb = kwargs.get('lb')
        if lb is None:
            lb = kwargs.get('lower_bound')
        ub = kwargs.get('ub')
        if ub is None:
            ub = kwargs.get('upper_bound')
        if lb is None:
            warnings.warn("No lower bound was supplied for random uniform distribution. Assuming lower bound of 0.")
            lb = 0.
        if ub is None:
            warnings.warn("No upper bound was supplied for random uniform distribution. Assuming upper bound of 1.")
            ub = 1.
        info = [lb, ub]
    elif distribution == 'random_normal':
        std = kwargs.get('std')
        if std is None:
            var = kwargs.get('var')
            if var is not None:
                std = np.sqrt(var)
        if std is None:
            warnings.warn("No standard deviation was supplied for random normal distribution. Assuming a standard "
                          "deviation of 1.")
            std = 1.
        info = std

    return distribution, seed, info



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from typing import Optional
import warnings

import casadi as ca

from ..base import Base, Equations
from ...plugins.plugins import LearningManager


class LearningBase(Base):
    """
    Base class for all machine learning classes

    :param features:
    :param labels:
    :param id:
    :param name:
    """
    def __init__(
            self,
            features: list[str],
            labels: list[str],
            id: Optional[str] = None,
            name: Optional[str] = None
    ) -> None:
        """Constructor method"""
        super().__init__(id=id, name=name)

        self._features = features
        self._n_features = len(features)
        self._labels = labels
        self._n_labels = len(labels)
        self._backend = None

    __array_ufunc__ = None  # https://stackoverflow.com/questions/38229953/array-and-rmul-operator-in-python-numpy

    def __matmul__(self, other):
        """Matrix multiplication method"""
        features = [ca.SX.sym(k) for k in self._features]
        out = self._function(features=ca.vertcat(*features))
        labels = out['labels']
        # NOTE: @-operator doesn't seem to work because we set __array_ufunc__ to None
        equations = Equations(ca.SX, {'matmul': ca.mtimes(labels, other)})
        return equations.to_function(f'{type(self).__name__}_matmul', *features)

    def __rmatmul__(self, other):
        """Matrix multiplication method (from the right side)"""
        features = [ca.SX.sym(k) for k in self._features]
        out = self._function(features=ca.vertcat(*features))
        labels = out['labels']
        # NOTE: @-operator doesn't seem to work because we set __array_ufunc__ to None
        equations = Equations(ca.SX, {'matmul': ca.mtimes(other, labels)})
        return equations.to_function(f'{type(self).__name__}_matmul', *features)

    def _set_backend(self, train_backend):
        """

        :param train_backend:
        :return:
        """
        if isinstance(train_backend, str):
            backend = train_backend.lower()
            available_backends = ['pytorch', 'tensorflow']
            if backend in available_backends:
                self._backend = LearningManager(backend)
            else:
                # TODO: Throw an error here?
                warnings.warn(f"Supplied backend '{train_backend}' for training neural networks is not supported. "
                              f"Either use your own backend or choose one of the following: "
                              f"{','.join(available_backends)}")
        else:
            self._backend = train_backend

    @property
    def features(self) -> list[str]:
        """

        :return:
        """
        return self._features

    @property
    def n_features(self) -> int:
        """

        :return:
        """
        return self._n_features

    @property
    def labels(self) -> list[str]:
        """

        :return:
        """
        return self._labels

    @property
    def n_labels(self) -> int:
        """

        :return:
        """
        return self._n_labels

    def update(self, *args):
        """

        :param args:
        :return:
        """
        pass



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

from abc import ABCMeta, abstractmethod
from copy import deepcopy
from typing import Optional, Sequence, TypeVar, Union
import warnings

import casadi as ca
import numpy as np

from .base import Base, Vector, Problem, OptimizationSeries, TimeSeries
from .dynamic_model.dynamic_model import Model
from ..util.optimizer import SciPyOptimizer as scisol
from ..util.util import check_and_wrap_to_list, check_if_list_of_string, dump_clean, generate_c_code, who_am_i, JIT

Numeric = Union[int, float]
NumArray = Union[Sequence[Numeric], np.ndarray]
ArrayLike = Union[list, tuple, dict, NumArray, ca.DM, Vector]
Symbolic = TypeVar('Symbolic', ca.SX, ca.MX)

SCIPY_OPTIMIZERS = ['CG', 'BFGS', 'Newton-CG', 'Nelder-Mead', 'L-BFGS-B', 'Powell']


class Optimizer(Base, metaclass=ABCMeta):
    """
    Base class for all optimizers

    :param id:
    :type id: str, optional
    :param name:
    :type name: str, optional
    :param solver:
    :type solver: str, optional
    :param solver_options:
    :type solver_options:
    :param plot_backend:
    :type plot_backend:
    """

    # TODO: Getters and setter for boxed constraints, inequality constraints and equality constraints
    # TODO: Remove methods
    # TODO: Add methods
    # TODO: Add solver check (similar to the one in the Model class)
    # TODO: Add parsing of optimization problem
    def __init__(
            self,
            id: Optional[str] = None,
            name: Optional[str] = None,
            solver: Optional[str] = None,
            solver_options: Optional[dict[str, Union[str, Numeric]]] = None,
            plot_backend: Optional[str] = None
    ) -> None:
        """Constructor method"""
        super().__init__(id=id, name=name)
        if self.id is None:
            self._create_id()

        prefix = self.name + '_' if self.name is not None else ''
        suffix = '_' + self._id.rsplit('_', 1)[-1]

        self._x = Vector(ca.SX, id='variables' + suffix, name=prefix + 'variables', parent=self)
        self._lbx = Vector(ca.DM, values_or_names=[], id='variables_lower_bound' + suffix,
                           name=prefix + 'variables_lower_bound', parent=self)
        self._ubx = Vector(ca.DM, values_or_names=[], id='variables_upper_bound' + suffix,
                           name=prefix + 'variables_upper_bound', parent=self)
        self._p = Vector(ca.SX, id='parameters' + suffix, name=prefix + 'parameters', parent=self)
        self._problem = Problem(parent=self)
        self._update_dimensions()

        if solver is None:
            solver = 'ipopt'
        if solver_options is None:
            solver_options = {}
        self._solver = solver
        self._solver_opts = solver_options

        self._is_linear = None

        hist = {}
        self._solution = OptimizationSeries(plot_backend, parent=self, **hist)

    def __getstate__(self) -> dict:
        # TODO: Test this
        state = self.__dict__.copy()
        # Remove unpicklable entries
        # del state['']
        # or make them picklable
        # state[''] = []
        return state

    def __setstate__(self, state: dict) -> None:
        # TODO: Test this
        self.__dict__.update(state)
        for attr in ['_x', '_p', '_problem']:  # TODO: Add constraints and objective
            pointer = getattr(self, attr)
            pointer.parent = self
        if not hasattr(self, 'name'):
            self.name = None

    def _check_bounds(self) -> None:
        """

        :return:
        """
        lbx, ubx = self.bounds
        diff = ubx - lbx
        mask = diff < 0
        indices = np.flatnonzero(mask)
        if indices.size > 0:
            warnings.warn(f"Inconsistencies in the variable bounds. Bounds flipped at the inconsistent positions. The "
                          f"lower bound was bigger than the upper bound at the following indices: "
                          f"{', '.join(str(k) for k in indices)}")
            lbx[indices], ubx[indices] = ubx[indices], lbx[indices]

    def _check_linearity(self) -> None:
        """

        :return:
        """
        is_linear = True
        for eq in ['obj', 'cons']:
            for var in ['_x', '_p']:
                if hasattr(self, var):
                    is_linear = ca.is_linear(getattr(self._problem, eq), getattr(self, var).values)
                if not is_linear:
                    break
            if not is_linear:
                break

        self._is_linear = is_linear

    def _update_dimensions(self) -> None:
        """

        :return:
        """
        self._n_x = self._x.size1()
        self._n_p = self._p.size1()
        self._n_o = self._problem.obj.size1()
        self._n_c = self._problem.cons.size1()

    @abstractmethod
    def _update_solver(self) -> None:
        """

        :return:
        """
        pass

    @property
    def decision_variables(self) -> Symbolic:
        """

        :return:
        """
        return self._x.values

    optimization_variables = decision_variables
    x = decision_variables

    # @decision_variables.setter
    def set_decision_variables(
            self,
            *args: str,
            lower_bound: Optional[Numeric] = None,
            upper_bound: Optional[Numeric] = None,
            description: Optional[Union[str, Sequence[str]]] = None,
            labels: Optional[Union[str, Sequence[str]]] = None,
            units: Optional[Union[str, Sequence[str]]] = None,
            **kwargs: str
    ) -> Symbolic:
        """

        :param args:
        :param lower_bound:
        :param upper_bound:
        :param description:
        :param labels:
        :param units:
        :param kwargs:
        :return:
        """
        if lower_bound is None:
            lower_bound = -ca.inf

        if upper_bound is None:
            upper_bound = ca.inf

        if description is None:
            kwargs['description'] = ''
        else:
            kwargs['description'] = description

        if labels is None:
            kwargs['labels'] = ''
        else:
            kwargs['labels'] = labels

        if units is None:
            kwargs['units'] = ''
        else:
            kwargs['units'] = units

        if len(args) == 1:
            if isinstance(args[0], int) and 'name' not in kwargs:
                kwargs['name'] = 'x'
            self._set('x', args[0], **kwargs)
        elif len(args) == 2:
            if isinstance(args[0], str) and isinstance(args[1], int):
                kwargs['n_dim'] = args[1]
                self._set('x', args[0], **kwargs)
            elif isinstance(args[0], int) and isinstance(args[1], str):
                kwargs['n_dim'] = args[0]
                self._set('x', args[1], **kwargs)
            else:
                self._set('x', args, **kwargs)
        elif len(args) > 2:
            self._set('x', args, **kwargs)
        else:
            self._set('x', None, **kwargs)

        self.bounds = (lower_bound, upper_bound)

        return self._x.values

    set_optimization_variables = set_decision_variables
    set_variables = set_decision_variables

    @property
    def lower_bound(self) -> ca.DM:
        """

        :return:
        """
        return self._lbx.values

    @lower_bound.setter
    def lower_bound(self, arg: Numeric) -> None:
        self._set('lbx', arg, shape=self._x.shape)

    lbx = lower_bound

    @property
    def upper_bound(self) -> ca.DM:
        """

        :return:
        """
        return self._ubx.values

    @upper_bound.setter
    def upper_bound(self, arg: Numeric) -> None:
        self._set('ubx', arg, shape=self._x.shape)

    ubx = upper_bound

    @property
    def bounds(self) -> (ca.DM, ca.DM):
        """

        :return:
        """
        return self._lbx.values, self._ubx.values

    @bounds.setter
    def bounds(self, args: Union[Numeric, (Numeric, Numeric)]) -> None:
        self.set_bounds(*args)

    def set_bounds(self, *args: Numeric, **kwargs: dict[str, Numeric]) -> None:
        """

        :param args:
        :param kwargs:
        :return:
        """
        if args and kwargs:
            raise TypeError("Optimizer.set_bounds() can only accept positional or keyword arguments, not both")

        lbx = self._lbx.values
        ubx = self._ubx.values

        if args:
            if len(args) == 1:
                lbx = args[0]
                ubx = args[0]
            elif len(args) == 2:
                lbx = args[0]
                ubx = args[1]
            else:
                raise ValueError(f"Too many input arguments for setting the boundaries. Only 2 arguments are expected, "
                                 f"got {len(args)}")

        if kwargs:
            lbx = kwargs.get('lower_bound', lbx)
            ubx = kwargs.get('upper_bound', ubx)

        self.lower_bound = lbx
        self.upper_bound = ubx
        self._check_bounds()

    @property
    def n_x(self) -> int:
        """

        :return:
        """
        return self._n_x

    @property
    def initial_guess(self) -> Optional[ca.DM]:
        """

        :return:
        """
        if 'x' not in self._solution or self._solution.get_by_id('x').is_empty():
            return None
        return self._solution.get_by_id('x:0')

    @initial_guess.setter
    def initial_guess(self, arg: Union[Numeric, ArrayLike]) -> None:
        self.set_initial_guess(arg)

    x0 = initial_guess

    def set_initial_guess(self, x0: Union[Numeric, ArrayLike]) -> None:
        """

        :param x0:
        :return:
        """
        if not self._solution.is_set_up():
            raise RuntimeError("Optimizer is not set up. Run Optimizer.setup() before setting the initial guess.")

        if not self._x.is_empty():
            # TODO: Maybe rewrite similar to Model.set_initial_conditions()
            self._solution.set('x', x0)
        else:
            # TODO: Raise an error here?
            warnings.warn("Initial guess cannot be set, since no decision variables are defined for the optimization "
                          "problem.")

    @property
    def parameters(self) -> Symbolic:
        """

        :return:
        """
        return self._p.values

    p = parameters

    # @parameters.setter
    def set_parameters(
            self,
            *args: str,
            description: Optional[Union[str, Sequence[str]]] = None,
            labels: Optional[Union[str, Sequence[str]]] = None,
            units: Optional[Union[str, Sequence[str]]] = None,
            **kwargs: str
    ) -> Symbolic:
        """

        :param args:
        :param description:
        :param labels:
        :param units:
        :param kwargs:
        :return:
        """
        if description is None:
            kwargs['description'] = ''
        else:
            kwargs['description'] = description

        if labels is None:
            kwargs['labels'] = ''
        else:
            kwargs['labels'] = labels

        if units is None:
            kwargs['units'] = ''
        else:
            kwargs['units'] = units

        if len(args) == 1:
            if isinstance(args[0], int) and 'name' not in kwargs:
                kwargs['name'] = 'p'
            self._set('p', args[0], **kwargs)
        elif len(args) == 2:
            if isinstance(args[0], str) and isinstance(args[1], int):
                kwargs['n_dim'] = args[1]
                self._set('p', args[0], **kwargs)
            elif isinstance(args[0], int) and isinstance(args[1], str):
                kwargs['n_dim'] = args[0]
                self._set('p', args[1], **kwargs)
            else:
                self._set('p', args, **kwargs)
        elif len(args) > 2:
            self._set('p', args, **kwargs)
        else:
            self._set('p', None, **kwargs)

        return self._p.values

    @property
    def n_p(self) -> int:
        """

        :return:
        """
        return self._n_p

    def set_parameter_values(self, p: Union[Numeric, ArrayLike]) -> None:
        """

        :param p:
        :return:
        """
        if not self._p.is_empty():
            self._solution.add('p', p)
        else:
            warnings.warn(f"The optimization problem {self.name} doesn't have any parameters")

    @property
    def objective(self) -> Symbolic:
        """

        :return:
        """
        return self._problem.obj

    @objective.setter
    def objective(self, arg: Symbolic) -> None:
        self.set_objective(arg)

    obj = objective

    def set_objective(self, arg: Symbolic) -> None:
        """

        :param arg:
        :return:
        """
        if isinstance(arg, int):
            raise TypeError(f"Wrong type of argument for function {who_am_i()}")
        else:
            self._set('obj', arg)

    @property
    def n_o(self) -> int:
        """

        :return:
        """
        return self._n_o

    @property
    def sense(self) -> str:
        """

        :return:
        """
        return self._problem.sense

    @sense.setter
    def sense(self, arg: str) -> None:
        self._problem.sense = arg

    @property
    def constraints(self) -> Symbolic:
        """

        :return:
        """
        return self._problem.cons

    cons = constraints

    # @cons.setter
    def set_constraints(self, *args, **kwargs) -> None:
        """

        :param args:
        :param kwargs:
        :return:
        """
        # TODO: Support for constraints in string format
        # TODO: Think of other ways to pass information
        self._set('cons', args, x=self._x.values, p=self._p.values)

    @property
    def n_c(self) -> int:
        """

        :return:
        """
        return self._n_c

    @property
    def solver(self) -> str:
        """

        :return:
        """
        return self._solver

    @solver.setter
    def solver(self, arg: str) -> None:
        if arg != self._solver:
            if isinstance(arg, str):
                if self.check_solver(arg):
                    self._solver = arg
                    self._solver_opts = {}
                    if self._display:
                        print("Solver options have been reset")
                else:
                    if self._display:
                        print("Solver not updated")
            else:
                raise TypeError("Solver type must be a string")

    @property
    def options(self) -> str:
        """

        :return:
        """
        return dump_clean(self._solver_opts)

    @property
    def solution(self) -> OptimizationSeries:
        """

        :return:
        """
        return self._solution

    def check_consistency(self) -> None:
        """

        :return:
        """
        if self._function is not None:
            if self._function.has_free():
                free_vars = ", ".join(self._function.get_free())
                msg = f"The optimization problem has the following free variables/parameters: {free_vars}"
                raise RuntimeError(msg)

        self._check_bounds()

    @abstractmethod
    def check_solver(self, solver: str) -> None:
        """

        :param solver:
        :return:
        """
        pass

    def setup(self, **kwargs) -> None:
        """

        :param kwargs:
        :return:
        """
        interface = kwargs.get('interface', None)
        if interface is None:
            interface = ca.nlpsol

        opts = kwargs.get('options', None)
        if opts is not None:
            options = deepcopy(opts)
        else:
            options = {}
        options.update(self._solver_opts)

        solver = self._problem.to_solver('solver', interface, options=options)

        use_c_code = kwargs.get('c_code', False)
        if use_c_code:
            gen_path, gen_name, gen_opts = self._generator(**kwargs)
            if gen_path is not None:
                self._c_name = generate_c_code(solver, gen_path, gen_name, opts=gen_opts)
                if self._compiler_opts['method'] in JIT and self._compiler_opts['compiler'] == 'shell':
                    self._compiler_opts['path'] = gen_path
                self._function = 'solver'
                super().setup()
            else:
                self._function = solver
        else:
            self._function = solver

        self.check_consistency()

        names = []
        vector = {}
        if self._x.names:
            names += ['x']
            vector['x'] = {
                'values_or_names': self._x.names,
                'description': self._x.description,
                'labels': self._x.labels,
                'units': self._x.units,
                'shape': (self._n_x, 0),
                'data_format': ca.DM
            }
        if self._p.names:
            names += ['p']
            vector['p'] = {
                'values_or_names': self._p.names,
                'description': self._p.description,
                'labels': self._p.labels,
                'units': self._p.units,
                'shape': (self._n_p, 0),
                'data_format': ca.DM
            }
        self._solution.setup(*names, **vector)

    def solve(self, *args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        # TODO: Store return_status and success (accessible via self._function.stats())
        if self._function is None:
            # NOTE: Not sure if we want to throw an error here
            raise RuntimeError("Optimizer is not set up. Run the method setup() before solving the problem.")

        # TODO: Need to rethink structure here
        warm_start = kwargs.get('warm_start')
        if warm_start is None:
            warm_start = False
        if warm_start:
            x0 = self._solution.get_by_id('x')
            if x0.size2() <= 1:
                warnings.warn("Warm start is not applicable, since no previous solution exists")
                if x0.is_empty():
                    self.set_initial_guess(x0=0.)
        else:
            if self._solution.get_by_id('x').is_empty():
                warnings.warn(f"No initial guess supplied. Using 0 as initial guess. Execute "
                              f"{self.__class__.__name__}.set_initial_guess(x0) to set a different initial guess.")
                self.set_initial_guess(self._n_x * [0.])

        p = kwargs.get('p')
        if p is not None:
            self._solution.add('p', p)
        elif self._solution.get_by_id('p').is_empty():
            warnings.warn(f"No parameter values supplied. Setting them to 0. Execute "
                          f"{self.__class__.__name__}.set_parameter_values(p) or supply them as a keyword argument to "
                          f"{self.__class__.__name__}.solve(...) to set different parameter values.")
            self.set_parameter_values(self._n_p * [0.])

        args = self._solution.get_function_args()
        args['lbx'] = self._lbx.values
        args['ubx'] = self._ubx.values
        # TODO: Niceify (don't access 'protected' attributes of self._problem here)
        args['lbg'] = self._problem._lbg
        args['ubg'] = self._problem._ubg

        result = self._function(**args)
        self._solution.update(**result)

    def stats(self) -> dict:
        """

        :return:
        """
        return self._function.stats()


class LinearProgram(Optimizer):
    """
    Linear programming problem

    :param id:
    :type id: str, optional
    :param name:
    :type name: str, optional
    :param solver:
    :type solver: str, optional
    :param solver_options:
    :type solver_options:
    :param plot_backend:
    :type plot_backend:
    """

    def __init__(
            self,
            id: Optional[str] = None,
            name: Optional[str] = None,
            solver: Optional[str] = None,
            solver_options: Optional[dict[str, Union[str, Numeric]]] = None,
            plot_backend: Optional[str] = None
    ) -> None:
        """Constructor method"""
        if solver is None:
            solver = 'clp'
        super().__init__(id=id, name=name, solver=solver, solver_options=solver_options, plot_backend=plot_backend)

    def _update_solver(self) -> None:
        """

        :return:
        """
        # TODO: Handle options in case of solver switches
        if isinstance(self._solver, str):
            if not ca.has_conic(self._solver) and not self._solver == 'ampl':
                if self._display:
                    print(f"Solver '{self._solver}' is either not available on your system or is not a suitable solver."
                          f"Switching to 'clp'...")
                self._solver = 'clp'
        else:
            self._solver = 'clp'

    def check_solver(self, solver: str) -> bool:
        """

        :param solver:
        :return:
        """
        if isinstance(solver, str):
            if ca.has_conic(solver):
                return True
            else:
                return False
        else:
            raise TypeError("Solver type must be a string")

    def setup(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        self._check_linearity()
        if not self._is_linear:
            raise RuntimeError("Linear programming was chosen as the optimizer, but the supplied problem is not linear")

        super().setup(interface=ca.qpsol, **kwargs)


class QuadraticProgram(LinearProgram):
    """"""


class NonlinearProgram(Optimizer):
    """
    Nonlinear programming problem

    :param id:
    :type id: str, optional
    :param name:
    :type name: str, optional
    :param solver:
    :type solver: str, optional
    :param solver_options:
    :type solver_options:
    :param plot_backend:
    :type plot_backend:
    """

    def __init__(
            self,
            id: Optional[str] = None,
            name: Optional[str] = None,
            solver: Optional[str] = None,
            solver_options: Optional[dict[str, Union[str, Numeric]]] = None,
            plot_backend: Optional[str] = None
    ) -> None:
        """Constructor method"""
        if solver is None:
            solver = 'ipopt'
        super().__init__(id=id, name=name, solver=solver, solver_options=solver_options, plot_backend=plot_backend)

    def _update_solver(self) -> None:
        """

        :return:
        """
        # TODO: Handle options in case of solver switches
        if isinstance(self._solver, str):
            if not ca.has_nlpsol(self._solver) and self._solver not in SCIPY_OPTIMIZERS and not self._solver == 'ampl':
                if self._display:
                    print(f"Solver '{self._solver}' is either not available on your system or is not a suitable solver."
                          f" Switching to 'ipopt'...")
                self._solver = 'ipopt'
        else:
            self._solver = 'ipopt'

    def check_solver(self, solver: str) -> bool:
        """

        :param solver:
        :return:
        """
        if isinstance(solver, str):
            if ca.has_nlpsol(solver) or solver in SCIPY_OPTIMIZERS:
                return True
            else:
                return False
        else:
            raise TypeError("Solver type must be a string")

    def setup(self, **kwargs):
        """

        :param kwargs:
        :return:
        """
        self._check_linearity()
        if self._is_linear:
            warnings.warn("Nonlinear programming was chosen as the optimizer, but the supplied problem is linear. "
                          "Switching to linear programming for this problem might result in a better performance.")

        if self._solver not in SCIPY_OPTIMIZERS:
            super().setup(interface=ca.nlpsol, **kwargs)
        else:
            super().setup(interface=scisol, **kwargs)


class DynamicOptimization(Base):
    """Base class for all MPC and MHE"""

    def __init__(self, model, id=None, name=None, plot_backend=None, stats=False, use_sx=True):
        """Constructor method"""
        super().__init__(id=id, name=name)
        if not isinstance(model, Model):
            raise TypeError('The model must be an object of the Model class.')

        if not model.is_setup():
            raise RuntimeError("Model is not set up. Run Model.setup() before passing it to the controller.")

        self._model = model.copy(setup=False, use_sx=use_sx)
        self._model_orig = model.copy(setup=False, use_sx=use_sx)
        self._sampling_interval = model.solution.dt

        # Solver status
        self._solver_status = 'init'

        # Flags
        self._nlp_setup_done = False
        self._box_constraints_is_set = False
        self._initial_guess_is_set = False
        self._nlp_options_is_set = False
        self._solver_options_is_set = False
        self._nlp_solver_is_set = False
        self._custom_constraint_flag = False
        self._custom_constraint_is_soft_flag = False
        self._change_mode_term_flat = False
        self._scaling_is_set = False
        self._time_varying_parameters_is_set = False
        self._sampling_time_is_set = False

        self._horizon = None
        # Options
        self.int_opts = []
        self.integration_method = []
        self.collocation_polynomials = []
        self.collocation_polynomials_degree = []
        self.collocation_polynomials_fin_elements = []
        self._nlp_opts = {}
        self._nlp_solution = None

        # Time varying parameters settings
        self._n_tvp = 0
        self._time_varying_parameters_horizon = ca.DM.zeros((0, 0))

        self._solution = TimeSeries(plot_backend, parent=self)

        self._solver_name_list_qp = ['qpoases', 'cplex', 'gurobi', 'oopq', 'sqic', 'nlp']
        self._solver_name_list_nlp = ['ipopt', 'bonmin', 'knitro', 'snopt', 'worhp', 'scpgen', 'sqpmethod', 'blocksqp',
                                      'AmplInterface']

        self._n_iterations = 0

        # Here flag for statistics
        self._stats = stats

    def _rearrange_parameters(self, tvp, cp):
        """
        Takes time-varying parameters, and not time-varying parameters and rearranges them in the correct order of the
        model parameter vector

        :param tvp:
        :param cp:
        :return:
        """
        if self._n_tvp != 0:
            p = ca.MX.sym('p', 0)
            p_names = self._model.parameter_names
            ii_tvp = 0
            ii_cp = 0
            for name in p_names:
                if name in self._time_varying_parameters:
                    p = ca.vertcat(p, tvp[ii_tvp])
                    ii_tvp += 1
                else:
                    p = ca.vertcat(p, cp[ii_cp])
                    ii_cp += 1
        else:
            p = cp

        return p

    def _rearrange_parameters_numeric(self, tvp, cp):
        """
        Takes time-varying parameters, and not time-varying parameters and rearranges them in the correct order of the
        model parameter vector

        :param tvp:
        :param cp:
        :return:
        """
        if self._n_tvp != 0:
            p = []
            p_names = self._model.parameter_names
            ii_tvp = 0
            ii_cp = 0
            for name in p_names:
                if name in self._time_varying_parameters:
                    p = ca.vertcat(p, tvp[ii_tvp])
                    ii_tvp += 1
                else:
                    p = ca.vertcat(p, cp[ii_cp])
                    ii_cp += 1
        else:
            p = cp

        return p

    def _print_message(self):
        """

        :return:
        """
        # TODO create solver statuses that are used by HILO such that we can use the same for different solvers
        if self._nlp_options['print_level'] == 0:
            pass
        if self._nlp_options['print_level'] == 1:
            if self._solver_name == 'ipopt':
                refer_ipopt = f"Please refer to ipopt documentation. For more info pass ipopt.print_level:5 to the " \
                              f"solver_options in the {self.__class__.__name__}.setup()"

                if self._solver_status_code == 1:
                    print(f"{self.__class__.__name__} found an optimal solution.")

                elif self._solver_status_code == 2:
                    print(f"{self.__class__.__name__} found an acceptable solution.")

                else:
                    print(f"Ups... {self.__class__.__name__} had some problems. The ipopt error message is: "
                          f"{self._solver_status}. " + refer_ipopt)

    def _populate_solution(self):
        """

        :return:
        """
        names = ['t']
        vector = {
            't': {
                'values_or_names': self._model._t.names,
                'description': self._model._t.description,
                'labels': self._model._t.labels,
                'units': self._model._t.units,
                'shape': (1, 0),
                'data_format': ca.DM
            }
        }
        if self._model_orig._x.names:
            names += ['x']
            vector['x'] = {
                'values_or_names': self._model_orig._x.names,
                'description': self._model_orig._x.description,
                'labels': self._model_orig._x.labels,
                'units': self._model_orig._x.units,
                'shape': (self._model_orig._n_x, 0),
                'data_format': ca.DM
            }
        if self._model_orig._y.names:
            names += ['y']
            vector['y'] = {
                'values_or_names': self._model_orig._y.names,
                'description': self._model_orig._y.description,
                'labels': self._model_orig._y.labels,
                'units': self._model_orig._y.units,
                'shape': (self._model_orig._n_y, 0),
                'data_format': ca.DM
            }
        if self._model_orig._z.names:
            names += ['z']
            vector['z'] = {
                'values_or_names': self._model_orig._z.names,
                'description': self._model_orig._z.description,
                'labels': self._model_orig._z.labels,
                'units': self._model_orig._z.units,
                'shape': (self._model_orig._n_z, 0),
                'data_format': ca.DM
            }
        if self._model_orig._u.names:
            names += ['u']
            vector['u'] = {
                'values_or_names': self._model_orig._u.names,
                'description': self._model_orig._u.description,
                'labels': self._model_orig._u.labels,
                'units': self._model_orig._u.units,
                'shape': (self._model_orig._n_u, 0),
                'data_format': ca.DM
            }
        if self._model_orig._p.names:
            names += ['p']
            vector['p'] = {
                'values_or_names': self._model_orig._p.names,
                'description': self._model_orig._p.description,
                'labels': self._model_orig._p.labels,
                'units': self._model_orig._p.units,
                'shape': (self._model_orig._n_p, 0),
                'data_format': ca.DM
            }
        if self._model_orig._n_q > 0:
            names += ['q']
            vector['q'] = {
                'values_or_names': 'q',
                'shape': (self._model_orig._n_q, 0),
                'data_format': ca.DM
            }

        if hasattr(self, 'type') and self.type == 'NMPC':
            names += [f'thetapfo']
            vector[f'thetapfo'] = {
                'values_or_names': [f'thetapfo{i + 1}' for i in range(self.n_of_path_vars)],
                'description': [f'theta p.f. num. {i + 1}' for i in range(self.n_of_path_vars)],
                'shape': (self.n_of_path_vars, 0),
                'data_format': ca.DM
            }

        if self._stats:
            names += ['extime']
            vector['extime'] = {
                'values_or_names': 'extime',
                'units': 'seconds',
                'shape': (1, 0),
                'data_format': ca.DM
            }
            names += ['niterations']
            vector['niterations'] = {
                'values_or_names': 'niterations',
                'shape': (1, 0),
                'data_format': ca.DM
            }
            names += ['solvstatus']
            vector['solvstatus'] = {
                'values_or_names': 'solvstatus',
                'shape': (1, 0),
                'data_format': ca.DM
            }
        self._solution.setup(*names, **vector)

    def _solver_status_wrapper(self):
        """

        :return:
        """
        # TODO: consider other outputs
        if self._solver_name == 'ipopt':
            self._solver_status = self._solver.stats()['return_status'].lower()
            if self._solver_status == 'solve_succeeded':
                self._solver_status_code = 1
            elif self._solver_status == 'solved_to_acceptable_level':
                self._solver_status_code = 2
            elif self._solver_status == 'infeasible_problem_detected':
                self._solver_status_code = 3
            elif self._solver_status == 'restoration_failed':
                self._solver_status_code = 4
            elif self._solver_status == 'maximum_iterations_exceeded':
                self._solver_status_code = 5
            else:
                self._solver_status_code = -1

    def _get_tvp_parameters_values(self):
        """
        Given the current iteration, returns the value for every sampling time of all time-varying parameter

        :return:
        """
        ci = self._n_iterations

        # Shift the horizon of tvp one step back
        if ci > 0:
            self._time_varying_parameters_horizon[:, 0:-1] = self._time_varying_parameters_horizon[:, 1:]
            for k, name in enumerate(self._time_varying_parameters):
                value = self._time_varying_parameters_values[name]

                if ci + self.horizon > len(value):
                    warnings.warn("The prediction horizon is predicting outside the values of the time varying "
                                  "parameters. I am now taking looping back the values and start from there. "
                                  "See documentation for info.")

                    n_of_overtakes = int(np.floor(ci / self.horizon))

                    self._time_varying_parameters_horizon[k, -1] = value[
                        ci - self.horizon * n_of_overtakes]
                else:
                    self._time_varying_parameters_horizon[k, -1] = value[ci + self.horizon - 1]
        else:
            self._time_varying_parameters_horizon = ca.DM.zeros((self._n_tvp), self.prediction_horizon)
            tvp_counter = 0
            for key, value in self._time_varying_parameters_values.items():
                if len(value) < self.prediction_horizon:
                    raise TypeError(f"When passing time-varying parameters, you need to pass a number of values at "
                                    f"least as long as the prediction horizon. The parameter {key} has {len(value)} "
                                    f"values but the MPC has a prediction horizon length of {self._prediction_horizon}."
                                    )

                value = self._time_varying_parameters_values[key]
                self._time_varying_parameters_horizon[tvp_counter, :] = value[0:self._prediction_horizon]
                tvp_counter += 1

    def reset_solution(self) -> None:
        """

        :return:
        """
        if not self._solution.is_empty():
            for k in self._solution:
                self._solution.remove(k, slice(0, None))

    def set_stage_constraints(self, stage_constraint=None, lb=None, ub=None, is_soft=False, max_violation=ca.inf,
                              weight=None, name='stage_constraint'):
        """
        Allows to add a (nonlinear) stage constraint.

        :param stage_constraint: SX expression. It has to contain variables of the model
        :param lb: Vector or list float,integer or casadi.DM. Lower bound on the constraint
        :param ub: Vector or list float,integer or casadi.DM. Upper bound on the constraint
        :param is_soft: bool: if True soft constraints are used.
        :param max_violation: (optional) Vector or list float,integer or casadi.DM. Maximum violation if constraint is
            soft. If None, there is no limit on the violation of the soft constraints
        :param weight: (optional) matrix of appropriate dimension. If is_soft=True it will be used to weight the soft
            constraint in the objective function using a quadratic cost.
        :param name:
        :return:
        """
        # TODO check if all the variables used in the function are in the model
        # TODO allow to pass either only the lower or the upper bound
        self.stage_constraint.constraint = stage_constraint
        self.stage_constraint.lb = lb
        self.stage_constraint.ub = ub
        self.stage_constraint.is_soft = is_soft
        self.stage_constraint.max_violation = max_violation
        self.stage_constraint.weight = weight
        self.stage_constraint.name = name

    def set_custom_constraints_function(self, fun=None, lb=None, ub=None, soft=False, max_violation=ca.inf):

        """
        Set a custom function that will be added as last position in the nonlinear constraints vector:
                                            lb <= fun <= ub
        This must take the entire optimization vector z and the indices of states and inputs as
                                            fun(v, x_ind, u_ind)

        :param fun: python function
        :param lb: lower bound
        :param ub: upper bound
        :param soft:
        :param max_violation:
        :return:
        """
        if lb is None:
            lb = [-ca.inf]
        lb = check_and_wrap_to_list(lb)
        if ub is None:
            ub = [ca.inf]
        ub = check_and_wrap_to_list(ub)

        self._custom_constraint_fun = fun
        self._custom_constraint_fun_lb = lb
        self._custom_constraint_fun_ub = ub
        self._custom_constraint_size = len(lb)
        self._custom_constraint_flag = True
        self._custom_constraint_is_soft_flag = soft
        self._custom_constraint_maximum_violation = max_violation

    def set_box_constraints(self, x_ub=None, x_lb=None, u_ub=None, u_lb=None, y_ub=None, y_lb=None, z_ub=None,
                            z_lb=None):
        """
        Set box constraints to the model's variables. These look like

        .. math::
                                    x_{lb} \leq x \leq x_{ub}

        :param x_ub: upper bound on states.
        :type x_ub: list, numpy array or CasADi DM array
        :param x_lb: lower bound on  states
        :type x_lb: list, numpy array or CasADi DM array
        :param u_ub: upper bound on inputs
        :type u_ub: list, numpy array or CasADi DM array
        :param u_lb: lower bound on inputs
        :type u_lb: list, numpy array or CasADi DM array
        :param y_ub: upper bound on measurements
        :type y_ub: list, numpy array or CasADi DM array
        :param y_lb: lower bound on measurements
        :type y_lb: list, numpy array or CasADi DM array
        :param z_ub: upper bound on algebraic states
        :type z_ub: list, numpy array or CasADi DM array
        :param z_lb: lower bound on algebraic states
        :type z_lb: list, numpy array or CasADi DM array
        :return:
        """
        if x_ub is not None:
            x_ub = deepcopy(x_ub)
            x_ub = check_and_wrap_to_list(x_ub)
            if len(x_lb) != self._n_x:
                raise TypeError(f"The model has {self._n_x} states. You need to pass the same number of bounds.")
            self._x_ub = x_ub
        else:
            self._x_ub = [ca.inf for i in range(self._model.n_x)]

        if x_lb is not None:
            x_lb = deepcopy(x_lb)
            x_lb = check_and_wrap_to_list(x_lb)
            if len(x_lb) != self._n_x:
                raise TypeError(f"The model has {self._n_x} states. You need to pass the same number of bounds.")
            self._x_lb = x_lb
        else:
            self._x_lb = [-ca.inf for i in range(self._model.n_x)]

        # Input constraints
        if u_ub is not None:
            u_ub = deepcopy(u_ub)
            u_ub = check_and_wrap_to_list(u_ub)
            if len(u_ub) != self._n_u:
                raise TypeError(f"The model has {self._n_u} inputs. You need to pass the same number of bounds.")
            self._u_ub = u_ub
        else:
            self._u_ub = [ca.inf for i in range(self._model.n_u)]

        if u_lb is not None:
            u_lb = deepcopy(u_lb)
            u_lb = check_and_wrap_to_list(u_lb)
            if len(u_lb) != self._n_u:
                raise TypeError(f"The model has {self._n_u} inputs. You need to pass the same number of bounds.")
            self._u_lb = u_lb
        else:
            self._u_lb = [-ca.inf for i in range(self._model.n_u)]

        # Algebraic constraints
        if z_ub is not None:
            z_ub = deepcopy(z_ub)
            z_ub = check_and_wrap_to_list(z_ub)
            if len(z_ub) != self._n_z:
                raise TypeError(f"The model has {self._n_z} algebraic states. You need to pass the same number of "
                                f"bounds.")
        else:
            self._z_ub = [ca.inf for i in range(self._model.n_z)]

        if z_lb is not None:
            z_lb = deepcopy(z_lb)
            z_lb = check_and_wrap_to_list(z_lb)
            if len(z_lb) != self._n_z:
                raise TypeError(f"The model has {self._n_z} algebraic states. You need to pass the same number of "
                                f"bounds.")
        else:
            self._z_lb = [-ca.inf for i in range(self._model.n_z)]

        if y_lb is not None or y_ub is not None:
            # Measurement box constraints can be added by an extra stage and terminal constraint (possibly nonlinear)
            self.set_stage_constraints(stage_constraint=self._model.meas, ub=deepcopy(y_ub), lb=deepcopy(y_lb),
                                       name='measurement_constraint')
            self.set_terminal_constraints(terminal_constraint=self._model.meas, ub=deepcopy(y_ub), lb=deepcopy(y_lb),
                                          name='measurement_constraint')

        self._box_constraints_is_set = True

    def set_initial_guess(self, x_guess=None, u_guess=None, z_guess=None):
        """
        Sets initial guess for the optimizer when no other information of the states or inputs are available.
        :param x_guess: optimal state guesss
        :type x_guess: list, numpy array, float or casADi DM array
        :param u_guess: optimal input guess
        :type u_guess: list, numpy array, float or casADi DM array
        :param z_guess: optimal algebraic states guess
        :type z_guess: list, numpy array, float or casADi DM array
        :return:
        """
        # TODO: make this nicer
        if x_guess is not None:
            self._x_guess = check_and_wrap_to_list(deepcopy(x_guess))

            if len(self._x_guess) != self._model_orig.n_x:
                raise ValueError(f"x_guess dimension and model state dimension do not match. Model state has dimension "
                                 f"{self._model_orig.n_x} while x_guess has dimension {len(self._x_guess)}")
        else:
            self._x_guess = self._model_orig.n_x * [0]

        if u_guess is not None:
            self._u_guess = check_and_wrap_to_list(deepcopy(u_guess))

            if len(self._u_guess) != self._model_orig.n_u:
                raise ValueError(f"u_guess dimension and model u dimension do not match. Model u_guess has dimension "
                                 f"{len(self._u_guess)} while the model input has dimension {self._model.n_u}")
        else:
            self._u_guess = self._model_orig.n_u * [0]

        if z_guess is not None:
            self._z_guess = check_and_wrap_to_list(deepcopy(z_guess))

            if len(self._z_guess) != self._model_orig.n_z:
                raise ValueError(f"z_guess dimension and model z dimension do not match. Model z has dimension "
                                 f"{self._model_orig.n_z} while z_guess {len(self._z_guess)}")
        else:
            self._z_guess = self._model_orig.n_z * [0]

        self._initial_guess_is_set = True

    def set_solver_opts(self, *args, **kwargs):
        """

        Sets the nonlinear programming settings. This can be a dictionary or key-valued arguments
        These correspond to the option of the used nlp solver, for example ipopt or bonmin
        :param args:
        :param kwargs:
        :return:
        """
        if len(args) != 0:
            if isinstance(args[0], dict):
                self._nlp_opts = args[0]
        else:
            if len(kwargs) != 0:
                self._nlp_opts = kwargs[0]
            else:
                self._nlp_opts = {}

        if self._nlp_options_is_set and self._nlp_options['print_level'] == 1:
            if self._nlp_opts.get('ipopt.print_level') is None:
                self._nlp_opts['ipopt.suppress_all_output'] = 'yes'

            if self._nlp_opts.get('print_time') is None:
                self._nlp_opts['print_time'] = False
        else:
            self._nlp_opts['ipopt.suppress_all_output'] = 'yes'
            self._nlp_opts['print_time'] = False

        self._solver_options_is_set = True

    def set_nlp_solver(self, solver):
        """
        Set the solver for the MPC. Possible options are
        For nonlinear programming problems (e.g. nonlinear MPC)
        - 'ipopt', 'bonmin', 'knitro', 'snopt', 'worhp', 'scpgen', 'sqpmethod', 'blocksqp', 'AmplInterface'
        for quadratic programmin problems (e.g. linear MPC)
        - 'qpoases', 'cplex', 'gurobi', 'oopq', 'sqic', 'nlp
        For more info refer to the CasADi documentation https://web.casadi.org/docs/.

        :param solver:
        :return:
        """
        self._solver_name_list = self._solver_name_list_qp + self._solver_name_list_nlp
        self._solver_name = solver
        self._nlp_solver_is_set = True

    def set_nlp_options(self, *args, **kwargs):
        """
        Sets the options that modify how the MPC problem is set

        :param args:
        :param kwargs:
        :return:
        """
        # TODO: when multiple-shooting and irk are implemented/tested add them to the list
        possible_choices = {}
        possible_choices['integration_method'] = ['collocation', 'rk4', 'erk', 'discrete', 'idas', 'cvodes']  # 'irk'
        possible_choices['solver'] = self._solver_name_list_nlp
        possible_choices['collocation_points'] = ['radau', 'legendre']
        possible_choices['objective_function'] = ['discrete', 'continuous']
        possible_choices['warm_start'] = [True, False]
        possible_choices['degree'] = None
        possible_choices['print_level'] = [0, 1]
        possible_choices['ipopt_debugger'] = [True, False]

        if self._type == 'SMPC':
            possible_choices['chance_constraints'] = ['prs']

        default_opts = {
            'integration_method': 'collocation',
            'collocation_points': 'radau',
            'degree': 3,
            'print_level': 1,
            'warm_start': True,
            'solver': 'ipopt',
            'ipopt_debugger': False
        }

        if self._type == 'SMPC':
            default_opts.update({'chance_constraints':'prs'})

        if self._model.discrete:
            default_opts.update({'objective_function': 'discrete'})
        else:
            default_opts.update({'objective_function': 'continuous'})

        option_list = list(default_opts.keys())
        opts = {}
        if len(args) != 0:
            if isinstance(args[0], dict):
                opts = args[0]
        else:
            if len(kwargs) != 0:
                opts = kwargs

        for key, value in opts.items():
            if key not in option_list:
                raise ValueError(f"The option named {key} does not exist. Possible options are {option_list}.")
            if possible_choices[key] is not None and value not in possible_choices[key]:
                raise ValueError(f"The option {key} is set to value {value} but the only allowed values are "
                                 f"{possible_choices[key]}.")
            else:
                default_opts[key] = value

        if default_opts.get('integration_method') != 'discrete' and self._model.discrete is True:
            warnings.warn(f"The integration method is set to {default_opts.get('integration_method')} but I notice that"
                          f" the model is in discrete time. I am overwriting and using discrete mode.")
            default_opts['integration_method'] = 'discrete'

        # Integration methods. Those are necessary for the RungeKutta class
        if default_opts['integration_method'] == 'rk4':
            default_opts['class'] = 'explicit'
            default_opts['method'] = 'rk'
            default_opts['order'] = 4
            default_opts['category'] = 'runge-kutta'
        elif default_opts['integration_method'] == 'erk':
            default_opts['class'] = 'explicit'
            default_opts['method'] = 'rk'
            default_opts['category'] = 'runge-kutta'
        elif default_opts['integration_method'] == 'irk':
            default_opts['class'] = 'implicit'
            default_opts['method'] = 'rk'
            default_opts['category'] = 'runge-kutta'
        elif default_opts['integration_method'] == 'collocation':
            default_opts['class'] = 'implicit'
            default_opts['method'] = 'collocation'
            default_opts['category'] = 'runge-kutta'

        if not self._nlp_solver_is_set:
            self.set_nlp_solver(default_opts['solver'])

        self._nlp_options_is_set = True
        self._nlp_options = default_opts

    def set_scaling(self, x_scaling=None, u_scaling=None, y_scaling=None):
        """
        Pass the scaling factors of states, inputs and outputs.
        This is important for system with where the optimization variable can have a large difference of order of magnitude.
        The scaling factors divide the respective variables.

        :param x_scaling: scaling factors for the states
        :type x_scaling: list
        :param u_scaling: scaling of scaling factors for inputs
        :type u_scaling: list
        :param y_scaling: scaling of scaling factors for outputs
        :type y_scaling: list
        :return:
        """

        if x_scaling is None:
            self._x_scaling = [1 for i in range(self._model.n_x)]
        else:
            self._x_scaling = check_and_wrap_to_list(x_scaling)

        if u_scaling is None:
            self._u_scaling = [1 for i in range(self._model.n_u)]
        else:
            self._u_scaling = check_and_wrap_to_list(u_scaling)

        if y_scaling is None:
            self._y_scaling = [1 for i in range(self._model.n_y)]
        else:
            self._y_scaling = check_and_wrap_to_list(y_scaling)

        self._scaling_is_set = True

    def set_sampling_interval(self, dt=None):
        """

        :param dt:
        :return:
        """
        if dt is not None:
            if isinstance(dt, float) or isinstance(dt, int):
                self._sampling_interval = dt
            else:
                raise TypeError("Sampling interval must be a float.")

    def set_time_varying_parameters(self, names=None, values=None):
        """
        Sets the time-varying parameters values for then MPC.
        Time varying parameters are model parameters that change in time. Hence, when predicting the system dynamics,
        their value must be supplied to the model at the every sampling time.

        :param names: list of strings with time varying parameter names
        :type names: list of strings
        :param values: values of the time varying parameters. You need to pass at least a number of values as long as
            the prediction horizon. Note: if the prediction goes after the last supplied value, the values will be
            repeated!
        :type values: dict (optional)
        :return: None
        """

        if names is None:
            self._time_varying_parameters = []
        else:
            if not check_if_list_of_string(names):
                raise ValueError('Tvp must be a list of strings with the paramers name that are time varying')

            for k, tvp in enumerate(names):
                if tvp not in self._model.parameter_names:
                    raise ValueError(f"I could not find the parameter {tvp} in the model. The models parameters are "
                                     f"{self._model.parameter_names}.")
                else:
                    self._time_varying_parameters_ind.append(k)
            self._n_tvp = len(names)
            self._time_varying_parameters = names
        self._time_varying_parameters_is_set = True

        self._time_varying_parameters_values = values

        if values is not None:
            if not isinstance(values, dict):
                raise TypeError("The values parameter must be a dictionary.")
            # Check that the dictionary contains as key the names of the tvp:
            for key in values.keys():
                if key not in names:
                    raise ValueError(f"The key {key} is not in the name vector: {names}. You need to pass a dictionary "
                                     f"where the keys are the name of the time varying parameters.")

    def plot_iterations(self, plot_states=False, **kwargs):
        """
        This plots the states and constraints for every mpc iteration. Useful to debug.
        At the moment it is working only with Bokeh.
        Warning: if the optimizer computes a large number of iterations the plots might take a bit of time to show up.

        :param plot_states: if True, also the states are plotted
        :param kwargs: If you pass 'plot_last' only the last iteration is plotted. This is to avoid very busy figures.
        :return:
        """
        # TODO: expand this to matplotlib
        # TODO: allow usage in jupyter notebook
        # TODO: expand this to MHE
        import itertools

        from bokeh.layouts import column
        from bokeh.models import BoxAnnotation
        from bokeh.plotting import figure, show, gridplot
        from bokeh.palettes import inferno  # select a palette

        if not hasattr(self, 'debugger'):
            raise ValueError("You need to activate the debugger. Pass 'options={'ipopt_debugger': True}' in the setup "
                             "of the NMPC.")

        plot_last = kwargs.get('plot_last', False)
        numLines = len(self.debugger.g_sols)

        pg = figure(width=2000, height=1200, title='Nonlinear constraints')
        x = np.arange(0, self._g.shape[0])
        colors_g = itertools.cycle(inferno(numLines))
        if plot_last:
            y = self.debugger.g_sols[-1]
            pg.scatter(x, y.squeeze(), color=next(colors_g), legend_label=f'iter. {numLines}', line_width=2)
        else:
            for m in range(numLines):
                y = self.debugger.g_sols[m]
                pg.scatter(x, y.squeeze(), color=next(colors_g), legend_label=f'iter. {m}', line_width=2)

        # Add sections saying what is what
        for i in range(len(self._g_indices['dynamics_collocation'])):
            low_box = BoxAnnotation(left=self._g_indices['dynamics_collocation'][i][0],
                                    right=self._g_indices['dynamics_collocation'][i][1],
                                    fill_alpha=0.1, fill_color='red')
            pg.add_layout(low_box)

        for i in range(len(self._g_indices['dynamics_multiple_shooting'])):
            low_box = BoxAnnotation(left=self._g_indices['dynamics_multiple_shooting'][i][0],
                                    right=self._g_indices['dynamics_multiple_shooting'][i][1],
                                    fill_alpha=0.1, fill_color='blue')
            pg.add_layout(low_box)

        for i in range(len(self._g_indices['nonlin_stag_const'])):
            low_box = BoxAnnotation(left=self._g_indices['nonlin_stag_const'][i][0],
                                    right=self._g_indices['nonlin_stag_const'][i][1],
                                    fill_alpha=0.1, fill_color='green')
            pg.add_layout(low_box)

        for i in range(len(self._g_indices['nonlin_term_const'])):
            low_box = BoxAnnotation(left=self._g_indices['nonlin_term_const'][i][0],
                                    right=self._g_indices['nonlin_term_const'][i][1],
                                    fill_alpha=0.1, fill_color='orange')
            pg.add_layout(low_box)

        plagg = figure(width=2000, height=1200, title='Lagrange multipliers nonlinear constraints')
        x = np.arange(0, self._g.shape[0])
        colors_lagg = itertools.cycle(inferno(numLines))
        if plot_last:
            y = self.debugger.lam_g_sols[-1]
            plagg.line(x, y.squeeze(), color=next(colors_lagg), legend_label=f'iter. {numLines}')
        else:
            for m in range(numLines):
                y = self.debugger.lam_g_sols[m]
                plagg.line(x, y.squeeze(), color=next(colors_lagg), legend_label=f'iter. {m}')
        show(column([pg, plagg]))

        if plot_states:

            px = []
            for name in self._model.dynamical_state_names:
                p = figure(width=800, height=800, title=name)
                px.append(p)
            colors_x = itertools.cycle(inferno(numLines))
            if plot_last:
                color = next(colors_x)
                x_pred = np.zeros((self._model.n_x, self.prediction_horizon + 1))
                u_pred = np.zeros((self._model.n_u, self.control_horizon))
                dt_pred = np.zeros((self.prediction_horizon))

                for ii in range(self.prediction_horizon + 1):
                    x_pred[:, ii] = np.array(self.debugger.x_sols[-1])[self._x_ind[ii]] * self._x_scaling
                for ii in range(self.control_horizon):
                    u_pred[:, ii] = np.array(self.debugger.x_sols[-1])[self._u_ind[ii]] * self._u_scaling
                if len(self._dt_ind) > 0:
                    for ii in range(self.prediction_horizon):
                        dt_pred[ii] = np.array(self.debugger.x_sols[-1])[self._dt_ind[ii]]
                else:
                    dt_pred = np.arange(0, (self.prediction_horizon + 1) * self._sampling_interval,
                                        self._sampling_interval)

                for k in range(self._n_x):
                    px[k].line(dt_pred, x_pred[k, :], color=color, legend_label=f'iter. {numLines}')
            else:
                for m in range(numLines):
                    color = next(colors_x)
                    x_pred = np.zeros((self._model.n_x, self.prediction_horizon + 1))
                    u_pred = np.zeros((self._model.n_u, self.control_horizon))
                    dt_pred = np.zeros((self.prediction_horizon))

                    for ii in range(self.prediction_horizon + 1):
                        x_pred[:, ii] = np.array(self.debugger.x_sols[m])[self._x_ind[ii]] * self._x_scaling
                    for ii in range(self.control_horizon):
                        u_pred[:, ii] = np.array(self.debugger.x_sols[m])[self._u_ind[ii]] * self._u_scaling
                    if len(self._dt_ind) > 0:
                        for ii in range(self.prediction_horizon):
                            dt_pred[ii] = np.array(self.debugger.x_sols[m])[self._dt_ind[ii]]
                    else:
                        dt_pred = np.arange(0, (self.prediction_horizon + 1) * self._sampling_interval,
                                            self._sampling_interval)

                    for k in range(self._n_x):
                        px[k].line(dt_pred, x_pred[k, :], color=color, legend_label=f'iter. {m}')

            show(gridplot(px, ncols=3))

    @property
    def sampling_interval(self):
        """

        :return:
        """
        return self._sampling_interval

    @property
    def initial_time(self):
        """

        :return:
        """
        return self._time

    @initial_time.setter
    def initial_time(self, arg):
        self._time = arg

    @property
    def obj_function_stage_cost(self):
        """

        :return:
        """
        return self._lag_term

    @property
    def current_time(self):
        """

        :return:
        """
        return self._time

    @property
    def horizon(self):
        """

        :return:
        """
        return self._horizon

    @horizon.setter
    def horizon(self, arg):
        if isinstance(arg, int) and arg >= 1:
            self._horizon = arg
            self._prediction_horizon = arg
            self._control_horizon = arg
            self._prediction_horizon_is_set = True
            self._control_horizon_is_set = True
        else:
            raise ValueError("The horizon must be a positive nonzero integer")

    @property
    def solution(self):
        """

        :return:
        """
        return self._solution

    @property
    def n_iterations(self):
        """

        :return:
        """
        return self._n_iterations

    def is_setup(self):
        """

        :return:
        """
        # TODO: Could we use self._function instead of self._solver? Then we would not need to overwrite this function.
        return hasattr(self, '_solver')


__all__ = [
    'LinearProgram',
    'QuadraticProgram',
    'NonlinearProgram'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

import copy
from distutils.version import StrictVersion
from typing import Optional
import warnings

import tensorflow as tf
from tensorflow.keras import Model
from tensorflow.keras.layers import Dense, Dropout, Input
import tensorflow.keras.losses as tfkl
from tensorflow.keras.optimizers import Adadelta, Adagrad, Adam, Adamax, Ftrl, Nadam, RMSprop, SGD
import tensorflow.keras.metrics as tfkm
from tensorflow.keras.callbacks import CallbackList, EarlyStopping, ReduceLROnPlateau
if StrictVersion(tf.__version__) >= StrictVersion('2.5.0'):
    from tensorflow.python.keras.utils.tf_utils import sync_to_numpy_or_python_type
else:
    from tensorflow.python.keras.utils.tf_utils import to_numpy_or_python_type as sync_to_numpy_or_python_type

from ...util.machine_learning import net_to_casadi_graph


def _mlp(n_features, n_labels, layers):
    """

    :param n_features:
    :param n_labels:
    :param layers:
    :return:
    """
    hidden = _process_layers(layers)
    inputs = Input(shape=(n_features, ))  # add name?
    x = inputs
    for layer in hidden:
        x = layer(x)
    # NOTE: Right now the output layer is assumed to be linear and cannot be changed
    outputs = Dense(n_labels)(x)

    return Model(inputs=inputs, outputs=outputs)


LOSS = {
    'mse': tfkl.MeanSquaredError,
    'mae': tfkl.MeanAbsoluteError,
    'mape': tfkl.MeanAbsolutePercentageError,
    'msle': tfkl.MeanSquaredLogarithmicError,
    'cs': tfkl.CosineSimilarity,
    'huber': tfkl.Huber,
    'log-cosh': tfkl.LogCosh
}
OPTIMIZER = {
    'adadelta': Adadelta,
    'adagrad': Adagrad,
    'adam': Adam,
    # 'adamw': NotImplemented,
    # 'sparse_adam': NotImplemented,
    'adamax': Adamax,
    # 'asgd': NotImplemented,
    'ftrl': Ftrl,
    # 'lbfgs': NotImplemented,
    'nadam': Nadam,
    'rmsprop': RMSprop,
    # 'rprop': NotImplemented,
    'sgd': SGD
}
METRIC = {
    'mse': tfkm.MeanSquaredError,
    'rmse': tfkm.RootMeanSquaredError,
    'mae': tfkm.MeanAbsoluteError,
    'mape': tfkm.MeanAbsolutePercentageError,
    'msle': tfkm.MeanSquaredLogarithmicError,
    'cs': tfkm.CosineSimilarity,
    'log-cosh': tfkm.LogCoshError
}


class _TensorFlowWrapper:
    """"""
    def __init__(self, module, **kwargs):
        """Constructor method"""
        self._module = module

        self._seed = kwargs.get('seed')

        learning_rate = kwargs.get('learning_rate')
        if learning_rate is None:
            learning_rate = .001

        loss = kwargs.get('loss')
        if loss is None:
            loss = 'mse'
        loss = LOSS[loss.lower()]()
        self._loss = loss

        optimizer = kwargs.get('optimizer')
        if optimizer is None:
            optimizer = 'adam'
        optimizer = OPTIMIZER[optimizer.lower()](learning_rate=learning_rate)
        self._optimizer = optimizer

        metric = kwargs.get('metric')
        if metric is None:
            metric = []
        self._metrics = [METRIC[k.lower()]() for k in metric]
        self._metric_names = metric

        # self._module.compile(optimizer=optimizer, loss=loss)

        self._tensorboard = kwargs.get('tensorboard')
        self._tensorboard_comment = {
            'batch_size': 0,
            'learning_rate': learning_rate,
            'n_layers': len(self._module.layers) - 2,
            'nodes': tuple(layer.units for k, layer in enumerate(self._module.layers) if hasattr(layer, 'units'))[:-1],
            'seed': self._seed
        }
        self._browser = kwargs.get('browser')

        self._train_loader = None
        self._test_loader = None
        self._early_stopping = None
        self._scheduler = None
        self._callbacks = None
        self._writer = None

    def _get_data_loader(self, data, batch_size, shuffle):
        """

        :param data:
        :param batch_size:
        :return:
        """
        data_loader = tf.data.Dataset.from_tensor_slices(data)
        if shuffle:
            data_loader = data_loader.shuffle(len(data_loader), seed=self._seed,
                                              reshuffle_each_iteration=True).batch(batch_size)
        else:
            data_loader = data_loader.batch(batch_size)
        return data_loader

    def _preprocessing(
            self,
            data,
            validation_data,
            batch_size,
            verbose,
            patience,
            shuffle
    ):
        """

        :param data:
        :param validation_data:
        :param batch_size:
        :param verbose:
        :param patience:
        :param shuffle:
        :return:
        """
        self._train_loader = self._get_data_loader(data, batch_size, shuffle)
        # TODO: Does it make sense to shuffle validation data
        self._test_loader = self._get_data_loader(validation_data, batch_size, False)

        if patience is not None and len(self._test_loader) > 0:
            # TODO: Check and implement behavior shown in https://www.tensorflow.org/guide/keras/custom_callback
            self._early_stopping = EarlyStopping(monitor='val_loss', patience=patience, verbose=verbose)
            self._scheduler = ReduceLROnPlateau(monitor='val_loss', patience=int(patience / 2), verbose=verbose)
            self._module.optimizer = self._optimizer
        elif patience is not None and len(self._test_loader) == 0:
            warnings.warn("Early stopping is only available when validation data is supplied for the training of the "
                          "neural network")

        if self._tensorboard is not None:
            self._tensorboard_comment['batch_size'] = batch_size
            self._writer = self._tensorboard.get_callback(**self._tensorboard_comment)
            if self._browser is not None:
                self.show_tensorboard()

    def _train(self, verbose):
        """

        :param verbose:
        :return:
        """
        running_loss = 0.

        # TODO: Need a better variable name
        batches = len(self._train_loader)

        logs = None
        for idx, (x, y) in enumerate(self._train_loader):
            self._callbacks.on_train_batch_begin(idx)
            loss = self._train_step(x, y)
            running_loss += float(loss)
            logs = {'loss': running_loss / (idx + 1)}
            logs.update({metric.name: float(metric.result()) for metric in self._metrics})
            self._callbacks.on_train_batch_end(idx + 1, logs=logs)
            if verbose == 1:
                _progress_bar(idx + 1, batches, logs, 30, "=", "\r")
            if self._module.stop_training:
                break
        if verbose == 1:
            print()
        elif verbose == 2:
            _one_line_per_epoch(batches, logs)
        logs = sync_to_numpy_or_python_type(logs)
        return logs

    @tf.function
    def _train_step(self, x, y):
        """

        :param x:
        :param y:
        :return:
        """
        with tf.GradientTape() as tape:
            pred = self._module(x, training=True)
            loss = self._loss(y, pred)
        grads = tape.gradient(loss, self._module.trainable_weights)
        self._optimizer.apply_gradients(zip(grads, self._module.trainable_weights))
        for metric in self._metrics:
            metric.update_state(y, pred)
        return loss

    def _test(self, data_loader, callbacks, verbose):
        """

        :param data_loader:
        :param callbacks:
        :param verbose:
        :return:
        """
        running_loss = 0.

        # TODO: Need a better variable name
        batches = len(data_loader)

        logs = None
        for idx, (x, y) in enumerate(data_loader):
            callbacks.on_test_batch_begin(idx)
            loss = self._test_step(x, y)
            running_loss += float(loss)
            logs = {'loss': running_loss / (idx + 1)}
            logs.update({metric.name: float(metric.result()) for metric in self._metrics})
            callbacks.on_test_batch_end(idx + 1, logs=logs)
            if verbose == 1:
                _progress_bar(idx + 1, batches, logs, 30, "=", "\r")
        if verbose == 1:
            print()
        elif verbose == 2:
            _one_line_per_epoch(batches, logs)
        logs = sync_to_numpy_or_python_type(logs)
        return logs

    @tf.function
    def _test_step(self, x, y):
        """

        :param x:
        :param y:
        :return:
        """
        pred = self._module(x, training=False)
        loss = self._loss(y, pred)
        for metric in self._metrics:
            metric.update_state(y, pred)
        return loss

    def train(
            self,
            data,
            validation_data,
            batch_size,
            epochs,
            verbose=1,
            patience=None,
            shuffle=True
    ):
        """

        :param data:
        :param validation_data:
        :param batch_size:
        :param epochs:
        :param verbose:
        :param patience:
        :param shuffle:
        :return:
        """
        self._preprocessing(data, validation_data, batch_size, verbose, patience, shuffle)

        callbacks = [self._early_stopping, self._scheduler, self._writer]
        callbacks = [callback for callback in callbacks if callback is not None]
        self._callbacks = CallbackList(callbacks=callbacks, model=self._module, verbose=verbose, epochs=epochs,
                                       steps=len(self._train_loader))
        self._callbacks.on_train_begin()

        self._module.stop_training = False
        training_logs = None
        for epoch in range(epochs):
            if verbose > 0:
                print(f"Epoch {epoch + 1}/{epochs}")
            self._callbacks.on_epoch_begin(epoch)

            logs = self._train(verbose)
            if logs is None:
                raise ValueError("Dataset is empty")
            for metric in self._metrics:
                metric.reset_states()
            epoch_logs = copy.copy(logs)

            if len(self._test_loader) > 1:
                if verbose > 0:
                    print("Evaluate on validation data")
                val_logs = self.evaluate(callbacks=self._callbacks, verbose=verbose)
                for metric in self._metrics:
                    metric.reset_states()
                val_logs = {'val_' + name: val for name, val in val_logs.items()}
                epoch_logs.update(val_logs)

            self._callbacks.on_epoch_end(epoch, logs=epoch_logs)
            training_logs = epoch_logs
            if self._module.stop_training:
                break

        self._callbacks.on_train_end(logs=training_logs)

    def evaluate(
            self,
            data=None,
            batch_size=None,
            verbose=1,
            callbacks=None
    ):
        """

        :param data:
        :param batch_size:
        :param verbose:
        :param callbacks:
        :return:
        """
        if data is not None:
            data_loader = self._get_data_loader(data, batch_size, False)
        else:
            data_loader = self._test_loader

        if not isinstance(callbacks, CallbackList):
            callbacks = CallbackList(callbacks=callbacks, model=self._module, verbose=verbose, epochs=1,
                                     steps=len(data_loader))

        callbacks.on_test_begin()
        val_logs = self._test(data_loader, callbacks, verbose)
        callbacks.on_test_end(logs=val_logs)

        return val_logs

    def build_graph(self, x, layers, input_scaling=None, output_scaling=None):
        """

        :param x:
        :param layers:
        :param input_scaling:
        :param output_scaling:
        :return:
        """
        return net_to_casadi_graph(self, x, layers, input_scaling=input_scaling, output_scaling=output_scaling)

    def get_weights_and_bias(self):
        """

        :return:
        """
        weights = []
        bias = []

        for val in self._module.get_weights():
            if val.ndim == 2:
                weights.append(val.T)
            elif val.ndim == 1:
                bias.append(val)

        return weights, bias

    def show_tensorboard(self, browser: Optional[str] = None) -> None:
        """

        :param browser:
        :return:
        """
        if browser is None:
            browser = self._browser
        browser = browser.lower()

        if self._tensorboard is not None:
            self._tensorboard.show(browser)
        else:
            warnings.warn("No TensorBoard was saved. If you want to see a TensorBoard set the corresponding flag "
                          "during setup and train again.")

    def close_tensorboard(self):
        """

        :return:
        """
        if self._tensorboard is not None:
            self._tensorboard.close()


LAYER = {
    'dense': Dense,
    'dropout': Dropout
}


def _process_layers(layers):
    """

    :param layers:
    :return:
    """
    hidden = []
    for layer in layers:
        type_ = layer.type.lower()
        if type_ == 'dropout':
            hidden.append(LAYER[type_](layer.rate))
        else:
            hidden.append(LAYER[type_](layer.nodes, activation=layer.activation.lower()))
    return hidden


def _progress_bar(iteration, total, logs, length, fill, print_end):
    """

    :param iteration:
    :param total:
    :param logs:
    :param length:
    :param fill:
    :param print_end:
    :return:
    """
    # TODO: Migrate together with _progress_bar from PyTorch wrapper
    loss_metrics_string = ""
    for key, val in logs.items():
        loss_metrics_string += f" - {key}: {val:.4f}"
    filled_length = int(length * iteration // total)
    unfilled_length = length - filled_length
    bar = fill * filled_length + ">" * bool(unfilled_length) + "." * (unfilled_length - 1)
    print(f"\r{iteration}/{total} [{bar}] {loss_metrics_string}", end=print_end)


def _one_line_per_epoch(total, logs):
    """

    :param total:
    :param logs:
    :return:
    """
    loss_metrics_string = ""
    for key, val in logs.items():
        loss_metrics_string += f" - {key}: {val:.4f}"
    print(f"{total}/{total} {loss_metrics_string}")


MODULES = {
    'mlp': _mlp
}


def get_wrapper(kind, *args, **kwargs):
    """

    :param kind:
    :param args:
    :param kwargs:
    :return:
    """
    # TODO: Support for device and seed
    seed = kwargs.get('seed')
    tf.random.set_seed(seed)

    module = MODULES[kind.lower()](*args)

    return _TensorFlowWrapper(module, **kwargs)


__all__ = [
    'get_wrapper'
]



#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

try:
    from collections.abc import ValuesView  # Python 3.3+
except ImportError:  # pragma: no cover - very old Python fallback
    from collections import ValuesView  # type: ignore
import functools
import platform
import sys
from typing import Any, Callable, Optional, TypeVar, Tuple, cast
import warnings

import casadi as ca
import numpy as np
from scipy.sparse import issparse
from scipy.sparse.linalg import norm as sparse_norm

if platform.system() == 'Windows':
    from .windows import get_vcvars, WINDOWS_COMPILERS
elif platform.system() == 'Linux':
    from .unix import find_compiler, UNIX_COMPILERS


Function = TypeVar('Function', bound=Callable[..., Any])


JIT = ['jit', 'just-in-time']
AOT = ['aot', 'ahead-of-time']
TYPES = {
    ca.SX: ca.SX,
    ca.MX: ca.MX,
    ca.DM: ca.DM,
    np.ndarray: np.array
}


def setup_warning(function: Function) -> Function:
    """

    :param function:
    :return:
    """
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        if args[0].is_setup():
            warnings.warn(f"{args[0].__class__.__name__} {args[0].name} was already set up. Please run the setup "
                          f"method again to apply the changes and prevent strange behavior.")
        return function(*args, **kwargs)
    return cast(Function, wrapper)


def _convert_operators(*args):
    """

    :param args:
    :return:
    """
    ret = set()
    for arg in args:
        if arg == '<' or arg == '>':
            ret.add(ca.OP_LT)
        elif arg == '<=' or arg == '>=':
            ret.add(ca.OP_LE)
        elif arg == '==':
            ret.add(ca.OP_EQ)
        elif arg == '+':
            ret.add(ca.OP_ADD)
        elif arg == '-':
            ret.add(ca.OP_SUB)
        elif arg == '*':
            ret.add(ca.OP_MUL)
        elif arg == '/':
            ret.add(ca.OP_DIV)

    return tuple(ret)


def _get_shape(**kwargs):
    """

    :param kwargs:
    :return:
    """
    # TODO: Typing hints
    n_dim = kwargs.get('n_dim', None)
    shape = kwargs.get('shape', None)
    if n_dim is None and shape is None:
        shape = (1, 1)
    elif n_dim is not None:
        if shape is None:
            axis = kwargs.get('axis', None)
            if axis is None or axis == 0:
                shape = (n_dim, 1)
            elif axis == 1:
                shape = (1, n_dim)
            elif axis > 1:
                raise IndexError("Argument 'axis' is out of bounds for array of dimension 2")
            else:
                raise TypeError("Expected argument 'axis' of type {}, got {} instead".format(type(int), type(axis)))
        else:
            if isinstance(shape, (int, tuple)):
                if isinstance(shape, int):
                    axis = kwargs.get('axis', None)
                    if axis is None or axis == 0:
                        shape = (shape, 1)
                    elif axis == 1:
                        shape = (1, shape)
                    elif axis > 1:
                        raise IndexError("Argument 'axis' is out of bounds for array of dimension 2")
                    else:
                        raise TypeError("Expected argument 'axis' of type {}, got {} instead".format(type(int),
                                                                                                     type(axis)))
                elif isinstance(shape, tuple):
                    if len(shape) > 2:
                        raise IndexError("Argument 'shape' is out of bounds for array of dimension 2")
                warnings.warn("Argument 'ndim' ignored, since argument 'shape' was supplied.")
            else:
                raise TypeError("Expected argument 'shape' of type {} or {}, got {} instead"
                                .format(type(int), type(tuple), type(shape)))

    return shape


def _split_expression(fun, var, *args):
    """

    :param fun:
    :param var:
    :param args:
    :return:
    """
    work = fun.sz_w() * ['']
    n_var = var.numel()
    split_ops = _convert_operators(*args)
    split_expr = []
    for k in range(fun.n_instructions()):
        op = fun.instruction_id(k)
        o = fun.instruction_output(k)
        o0 = -1
        o1 = -1
        if o:
            o0 = o[0]
        if len(o) > 1:
            o1 = o[1]
        i = fun.instruction_input(k)
        i0 = -1
        i1 = -1
        if i:
            i0 = i[0]
        if len(i) > 1:
            i1 = i[1]
        if op in split_ops:
            work[o0] = [work[i0], work[i1]]
        elif op == ca.OP_CONST:
            work[o0] = fun.instruction_constant(k)
        elif op == ca.OP_INPUT:
            work[o0] = var[i0 * n_var + i1]
        elif op == ca.OP_OUTPUT:
            # work[o1] = work[i0]
            split_expr.append(work[i0])
        elif op == ca.OP_ADD:
            work[o0] = work[i0] + work[i1]
        elif op == ca.OP_SUB:
            work[o0] = work[i0] - work[i1]
        elif op == ca.OP_MUL:
            work[o0] = work[i0] * work[i1]
        elif op == ca.OP_DIV:
            work[o0] = work[i0] / work[i1]
        else:
            raise ValueError(f"Operation with id '{op}' is unknown or not implemented")

    return split_expr


def check_and_wrap_to_list(arg):
    """
    Check if the input is an np.array or a list.

    :param arg:
    :return:
    """
    if isinstance(arg, np.ndarray):
        arg_ = arg.squeeze().tolist()
        if isinstance(arg_, int) or isinstance(arg_, float):
            # if arg is a np,array with a single value the .tolist() method returns a integer or float.
            arg_ = [arg]
    elif isinstance(arg, list):
        arg_ = arg
    elif isinstance(arg, float) or isinstance(arg, int) or isinstance(arg, str):
        arg_ = [arg]
    elif isinstance(arg, ca.DM):
        arg_ = np.array(arg).squeeze(-1).tolist()
    else:
        raise TypeError(f"Type {type(arg)} not supported. Must be list,float,ca.DM or numpy array ")

    return arg_


def check_and_wrap_to_DM(arg):
    """

    :param arg:
    :return:
    """
    if isinstance(arg, np.ndarray):
        arg_ = ca.DM(arg.squeeze())
    elif isinstance(arg, list) or isinstance(arg, float) or isinstance(arg, int):
        arg_ = ca.DM(arg)
    elif isinstance(arg, ca.DM):
        arg_ = arg
    else:
        raise TypeError(f"Type {type(arg)} not supported. Must be list,float,ca.DM or numpy array ")
    return arg_


def check_compiler(method: str, compiler: str) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """

    :param method:
    :param compiler:
    :return:
    """
    method = lower_case(method)
    compiler = lower_case(compiler)
    if method in JIT:
        if ca.Importer.has_plugin(compiler):
            if compiler == 'shell' and platform.system() == 'Windows':
                vcvars = get_vcvars()
                return method, compiler, vcvars
            else:
                return method, compiler, None
    elif method in AOT:
        if platform.system() == 'Linux' and compiler in UNIX_COMPILERS:
            if find_compiler(compiler):
                return method, compiler, None
        elif platform.system() == 'Windows' and compiler in WINDOWS_COMPILERS:
            vcvars = get_vcvars()
            if vcvars:
                return method, compiler, vcvars
    return None, None, None


def check_if_has_duplicates(vector):
    """
    Checks if list contains duplicates

    :param vector:
    :return:
    """
    if not isinstance(vector, list):
        raise TypeError("Vector must be a list.")

    if len(vector) == len(set(vector)):
        return False
    else:
        return True


def check_if_list_of_type(a, types):
    """

    :param a:
    :param types:
    :return:
    """
    # TODO: Typing hints
    if not isinstance(a, list):
        return False
    else:
        return all(isinstance(item, types) for item in a)


def check_if_list_of_none(a):
    """
    Check if a list contains only None

    :param a:
    :return:
    """
    return all(item is None for item in a)


def check_if_list_of_string(a):
    """
    Check if a list contains only strings

    :param a:
    :return:
    """
    return all(isinstance(item, str) for item in a)


def check_if_square(arg):
    """

    :param arg: DM or NP matrix
    :return: True if matrix is square, False if not
    """
    if len(arg.shape) > 2:
        raise ValueError("The input can have at most two dimensions.")

    if len(arg.shape) == 1:
        return False
    if len(arg.shape) == 2:
        if arg.shape[0] == arg.shape[1]:
            return True
        else:
            return False


def convert(obj, _type, **kwargs):
    """

    :param obj:
    :param _type:
    :param kwargs:
    :return:
    """
    # TODO: Typing hints
    # TODO: Convoluted. Simplify if possible.
    # TODO: Add support for dict?
    # TODO: Is there a nice way to convert SX to MX and vice versa?
    # TODO: Return success flag and error message for processing outside of convert
    fx = TYPES[_type]
    if obj is None:
        return convert(fx(), _type, **kwargs)
    if isinstance(obj, _type):
        shape = _get_shape(**kwargs)
        # TODO: What if size is to be reduced to (1, 1)? Should throw an error. Would do nothing here.
        if shape != (1, 1):
            if _type in [ca.SX, ca.MX, ca.DM]:
                if obj.is_scalar():
                    return ca.repmat(obj, shape)
                elif obj.is_column() or obj.is_row():
                    if max(obj.shape) != max(shape):
                        raise IndexError("Dimension mismatch. Cannot reduce or increase size of the shape.")
                    if obj.shape == shape[::-1]:
                        return obj.T
                    else:
                        raise Exception("Don't know how I got here. Please inform the maintainer.")
                elif obj.is_empty():
                    return obj.reshape(shape)
            else:
                if obj.size == 1:
                    return np.tile(obj, shape)
                elif obj.ndim == 1:
                    if len(shape) == 2 and shape[0] == shape[1] and shape[0] in obj.shape:
                        obj = np.diag(obj)
                elif 1 in obj.shape:
                    if max(obj.shape) != max(shape):
                        raise IndexError("Dimension mismatch. Cannot reduce or increase size of the shape.")
                    if obj.shape == shape[::-1]:
                        return obj.T
                    elif obj.ndim == 2:
                        if len(shape) == 2 and shape[0] == shape[1] and shape[0] in obj.shape:
                            obj = np.diag(obj.flatten())
                    else:
                        raise Exception("Don't know how I got here. Please inform the maintainer.")
        elif _type is np.ndarray and 'shape' in kwargs and obj.shape != shape:
            # TODO: Catch exceptions to generate meaningful error messages
            obj = np.reshape(obj, shape)
        return obj
    elif isinstance(obj, (list, tuple, np.ndarray, ca.DM)):
        # TODO: Add support for axis argument
        if isinstance(obj, (np.ndarray, ca.DM)):
            return convert(fx(obj), _type, **kwargs)
        elif not obj:
            return convert(fx(), _type, **kwargs)
        else:
            types = [isinstance(k, (_type, str, int, float)) for k in obj]
            not_supported = [k for k, val in enumerate(types) if not val]
            for k in not_supported:
                if hasattr(obj[k], 'values'):
                    obj[k] = obj[k].values
                    types[k] = True
            if all(types):
                if all(isinstance(k, _type) for k in obj):
                    return convert(ca.vertcat(*obj), _type, **kwargs)
                elif all(isinstance(k, str) for k in obj):
                    return convert([fx.sym(parse(k)) for k in obj], _type, **kwargs)
                elif all(isinstance(k, float) for k in obj):
                    return convert(fx(obj), _type, **kwargs)
                elif all(isinstance(k, int) for k in obj):
                    if isinstance(obj, tuple):
                        if len(obj) > 2:
                            raise IndexError("Shape is out of bounds for array of dimension 2")
                    elif isinstance(obj, list):
                        return convert(fx(obj), _type, **kwargs)
                    else:
                        raise TypeError("Expected argument of type {}, got {} instead".format(type(tuple), type(obj)))
                    name = kwargs.get('name', None)
                    if name is not None:
                        if isinstance(name, str):
                            return convert(fx.sym(name, obj), _type, **kwargs)
                        else:
                            raise TypeError("Expected argument 'name' of type {}, got {} instead"
                                            .format(type(str), type(name)))
                    else:
                        raise KeyError("Additional argument 'name' is missing")
                else:
                    # Mix of strings and SX/MX and possibly floats
                    res = []
                    for k in obj:
                        if isinstance(k, _type):
                            res.append(k)
                        else:
                            if isinstance(k, str):
                                res.append(fx.sym(k))
                            elif isinstance(k, (int, float)):
                                res.append(fx(k))
                            else:
                                raise TypeError(f"Wrong type of arguments for supplied list: {type(k).__name__}")
                    return convert(res, _type, **kwargs)
            else:
                raise TypeError("Wrong type of arguments for function {}".format(who_am_i()))
    elif isinstance(obj, str):
        shape = _get_shape(**kwargs)
        return fx.sym(parse(obj), shape)
    elif isinstance(obj, int):
        if _type in [ca.DM, np.ndarray]:
            return convert(fx(obj), _type, **kwargs)
        else:
            axis = kwargs.get('axis', None)
            if axis is None or axis == 0:
                shape = (obj, 1)
            elif axis == 1:
                shape = (1, obj)
            elif axis > 1:
                raise IndexError("Argument 'axis' is out of bounds for array of dimension 2")
            else:
                raise TypeError("Expected argument 'axis' of type {}, got {} instead".format(type(int), type(axis)))
            name = kwargs.get('name', None)
            if name is not None:
                if isinstance(name, str):
                    return fx.sym(name, shape)
                else:
                    raise TypeError("Expected argument 'name' of type {}, got {} instead".format(type(str), type(name)))
            else:
                raise KeyError("Additional argument 'name' is missing")
    elif isinstance(obj, float):
        return convert(fx(obj), _type, **kwargs)
    elif isinstance(obj, ValuesView):
        return convert(list(obj), _type, **kwargs)
    else:
        raise TypeError("Wrong type of arguments for function {}".format(who_am_i()))


def dump_clean(obj):
    """

    :param obj:
    :return:
    """
    # TODO: Typing hints
    # FIXME: Strings also have the attribute __iter__
    if isinstance(obj, dict):
        for k, v in obj.items():
            if hasattr(v, '__iter__'):
                print(k)
                dump_clean(v)
            else:
                print("{} : {}".format(k, v))
    elif isinstance(obj, list):
        for v in obj:
            if hasattr(v, '__iter__'):
                dump_clean(v)
            else:
                print(v)
    else:
        print(obj)


def generate_c_code(functions, path, name, opts=None):
    """

    :param functions:
    :param path:
    :param name:
    :param opts:
    :return:
    """
    if functions:
        if opts is None:
            opts = {}
        generator = ca.CodeGenerator(name, opts)
        if isinstance(functions, list):
            for f in functions:
                generator.add(f)
        else:
            generator.add(functions)
        c_name = generator.generate(path)
        return c_name
    else:
        return None


def is_array_like(obj):
    """

    :param obj:
    :return:
    """
    # TODO: Typing hints
    return is_list_like(obj) and hasattr(obj, 'dtype')


def is_diagonal(obj):
    """

    :param obj:
    :return:
    """
    diag = np.zeros(obj.shape)
    np.fill_diagonal(diag, obj.diagonal())
    return np.all(obj == diag)


def is_integer(n):
    """

    :param n:
    :return:
    """
    # TODO: Typing hints
    if isinstance(n, int):
        return True
    if isinstance(n, float):
        return n.is_integer()
    return False


def is_iterable(obj) -> bool:
    """

    :param obj:
    :return:
    """
    try:
        iter(obj)
    except TypeError:
        return False
    return True


def is_list_like(obj):
    """

    :param obj:
    :return:
    """
    # TODO: Typing hints
    return isinstance(obj, (list, tuple, set, np.ndarray))


def is_pd(x):
    """
    https://stackoverflow.com/a/63911811

    :param x:
    :return:
    """
    try:
        np.linalg.cholesky(x)
    except np.linalg.LinAlgError:
        return False
    return True


def is_psd(x, tol=1e-14):
    """
    https://stackoverflow.com/a/63911811

    :param x:
    :param tol:
    :return:
    """
    try:
        regularized_x = x + np.eye(x.shape[0]) * tol
        np.linalg.cholesky(regularized_x)
    except np.linalg.LinAlgError:
        return False
    return True


def is_real(x):
    """

    :param x:
    :return:
    """
    if not is_list_like(x):
        return np.isreal(x)
    else:
        return np.isreal(x).all()


def is_square(array):
    """

    :param array:
    :return:
    """
    if array.ndim == 1:
        return False
    return all([len(row) == len(array) for row in array])


def is_symmetric(x, tol=1e-8):
    """
    https://stackoverflow.com/a/67406237

    :param x:
    :param tol:
    :return:
    """
    if issparse(x):
        # TODO: Test this
        return sparse_norm(x - x.T, ca.inf) < tol
    else:
        return np.linalg.norm(x - x.T, ca.inf) < tol


def lower_case(obj: Any) -> Any:
    """

    :param obj:
    :return:
    """
    if isinstance(obj, dict):
        return {key.lower(): lower_case(val) for key, val in obj.items()}
    elif isinstance(obj, (list, set, tuple)):
        return type(obj)(lower_case(k) for k in obj)
    elif isinstance(obj, str):
        return obj.lower()
    else:
        return obj


def parse(string):
    """

    :param string:
    :return:
    """
    # TODO: Typing hints
    # TODO: Parse string, e.g. if equations are given in strings (could be useful later for writing equations in GUI)
    # Check if any of the strings in 'OPERATORS' is in 'string'
    return string


def random_state(state=None):
    """

    :param state:
    :return:
    """
    # TODO: Typing hints
    if is_integer(state) or is_array_like(state):
        return np.random.RandomState(state)
    elif isinstance(state, np.random.RandomState):
        return state
    elif state is None:
        return np.random
    else:
        raise ValueError("Argument 'random_state' must be an integer, array-like, a NumPy RandomState, or None")


def scale_vector(vector, scaler):
    """

    :param vector: NP array, list or ca.DM vector
    :param scaler:
    :return:
    """

    vector = check_and_wrap_to_list(vector)
    scaler = check_and_wrap_to_list(scaler)

    if len(vector) != len(scaler):
        raise ValueError(f"The length of scaler and vector must be equal. The vector has length {len(vector)} while the"
                         f" scaler {len(scaler)}")
    s_vector = vector.copy()
    for k, i in enumerate(vector):
        s_vector[k] = i / scaler[k]

    return s_vector


def who_am_i() -> str:
    """

    :return:
    """
    return sys._getframe(1).f_code.co_name


def clip(v,v_lb,v_ub):
    """
    Given a vector v, it returns the value of the vector clipped between v_lb <= v <=v_ub
    :param v: NP array, list or ca.DM vector
    :param v_lb: NP array, list or ca.DM vector
    :param v_ub: NP array, list or ca.DM vector
    :return: clipped vector v (casadi DM)
    """

    v = check_and_wrap_to_list(v)
    v_lb = check_and_wrap_to_list(v_lb)
    v_ub = check_and_wrap_to_list(v_ub)

    return ca.DM([v_lb[i] if v[i] < v_lb[i] else v_ub[i] if v[i] > v_ub[i] else v[i] for i in range(len(v))])


#   
#   This file is part of HILO-MPC
#
#   HILO-MPC is a toolbox for easy, flexible and fast development of machine-learning-supported
#   optimal control and estimation problems
#
#   Copyright (c) 2021 Johannes Pohlodek, Bruno Morabito, Rolf Findeisen
#                      All rights reserved
#
#   HILO-MPC is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Lesser General Public License as
#   published by the Free Software Foundation, either version 3
#   of the License, or (at your option) any later version.
#
#   HILO-MPC is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with HILO-MPC. If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import annotations

import os
from subprocess import Popen, PIPE, STDOUT
import sys
from typing import Optional


LETTERS = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
WINDOWS_COMPILERS = ['cl']


def compile_dll(path_to_file: str, output: Optional[str] = None) -> Optional[str]:
    """

    :param path_to_file:
    :param output:
    :return:
    """
    if os.path.isfile(path_to_file):
        vcvars = get_vcvars()
        if vcvars is not None:
            cl = f'cl.exe /LD {path_to_file}'
            if output is not None:
                cl += f' /F {output}'
            commands = ' && '.join((vcvars, cl, 'exit 0'))
            commands = f'({commands}) || exit 1\r\n'
            out = execute_commands(commands, encoding='latin1')
            if out is not None:
                if output is not None:
                    dll_path = output
                else:
                    dll_path = path_to_file.rsplit('.', 1)
                    dll_path[-1] = 'dll'
                return '.'.join(dll_path)
            else:
                return None
        else:
            return None
    else:
        raise FileNotFoundError(f"File {path_to_file} does not exist")


def execute_commands(commands: str, encoding: Optional[str] = None, errors: str = 'strict') -> str:
    """

    :param commands:
    :param encoding:
    :param errors:
    :return:
    """
    text_mode = (encoding is None)
    with Popen('cmd.exe', stdin=PIPE, stdout=PIPE, stderr=STDOUT, universal_newlines=text_mode) as process:
        if not text_mode:
            commands = commands.encode(encoding, errors)
        out, _ = process.communicate(commands)
    # Somehow err (2. output of communicate()) can be not empty although everything worked and is also not really
    # storing errors
    # NOTE: Workaround by using 'exit 0' and 'exit 1' in 'cmd'
    if process.returncode != 0:
        if not text_mode:
            out = out.decode(encoding, errors)
        raise RuntimeError(f"Could not execute commands"
                           f"{out}"
                           f"Return value: {process.returncode}")
    return out if text_mode else out.decode(encoding, errors)


def find_files(name: str) -> (list[str], list[str]):
    """

    :param name:
    :return:
    """
    loc = []
    date = []
    drives = get_hard_drives()
    for drive in drives:
        for root, dirs, files in os.walk(drive):
            for file in files:
                if file == name:
                    path_to_file = root + '\\' + file
                    loc.append(path_to_file)
                    date.append(os.path.getctime(path_to_file))
    if len(loc) != len(date):
        raise ValueError("List 'loc' and 'list' date must be of same length")
    return loc, date


def get_hard_drives() -> list[str]:
    """

    :return:
    """
    return [f'{drive}:\\' for drive in LETTERS if os.path.exists(f'{drive}:\\')]


def get_vcvars() -> str:
    """

    :return:
    """
    files, dates = find_files('vcvarsall.bat')
    arch = '64bit' if sys.maxsize > 2 ** 32 else '32bit'
    if files:
        index_max = max(range(len(dates)), key=dates.__getitem__)
        chosen = files[index_max]
        if ' ' in chosen:
            if arch == '64bit':
                return f'"{chosen}" x64'
            else:
                return f'"{chosen}" x32'
        else:
            if arch == '64bit':
                return f'{chosen} x64'
            else:
                return f'{chosen} x32'
    else:
        raise RuntimeError("Compiler not found")
